from dataclasses import dataclass
from enum import IntEnum
from typing import Union


class ShapeEnum(IntEnum):
    """
    The different simple shapes that can be created.
    """
    kBox = 1
    kSphere = 2
    kCylinder = 3
    kCapsule = 4

@dataclass
class SphereDefinition:
    radius: float
    type: ShapeEnum = ShapeEnum.kSphere


# Shape Definition Can be any one of these
ShapeDefinition = Union[SphereDefinition]


"""
Description:
    This file defines a Role class that is used to populate the
    scene with ONLY systems that match the appropriate structure/signature.
"""
from dataclasses import dataclass
from typing import List

from pydrake.systems.framework import DiagramBuilder

from brom_drake.scenes.roles.role_port_assignment import RolePortAssignment
# Internal Imports
from brom_drake.utils import (
    find_all_systems_with_output_port, find_all_systems_with_input_port,
    Performer,
)

# Class

@dataclass
class Role:
    name: str
    description: str
    port_assignments: List[RolePortAssignment]

    def __str__(self):
        return self.name

    def connect_performer_ports_to(
        self,
        builder: DiagramBuilder,
        performer: Performer,
    ):
        # Setup

        # Add performer to builder
        builder.AddSystem(performer)

        # Connect performer to the rest of the scene (stored in Builder)
        for assignment in self.port_assignments:
            assignment.create_connections_in(builder, performer)


"""
base.py
"""
from dataclasses import dataclass
from typing import Tuple, Callable

import networkx as nx
import numpy as np
from pydrake.geometry import SceneGraph
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex
# Internal Imports
from brom_drake.motion_planning.algorithms.motion_planner import MotionPlanner

@dataclass
class BaseRRTPlannerConfig:
    steering_step_size: float = 0.1
    prob_sample_goal: float = 0.075
    max_iterations: int = int(1e4)
    convergence_threshold: float = 1e-3

class BaseRRTPlanner(MotionPlanner):
    def __init__(
        self,
        robot_model_idx: ModelInstanceIndex,
        plant: MultibodyPlant,
        scene_graph: SceneGraph,
        config: BaseRRTPlannerConfig = None,
    ):
        super().__init__(robot_model_idx, plant, scene_graph)
        # Input Processing

        # Setup
        # self.dim_q = plant.num_actuated_dofs(robot_model_idx)

        self.config = config
        if self.config is None:
            self.config = BaseRRTPlannerConfig()

        # Prepare for planning
        # self.joint_limits = self.get_joint_limits()  # Initialize joint limits array

    @property
    def dim_q(self) -> int:
        if not self.plant.is_finalized():
            raise ValueError("Plant has not been finalized yet! Can not compute num_actuated_dofs().")

        return self.plant.num_actuated_dofs(self.robot_model_idx)

    @property
    def joint_limits(self) -> np.ndarray:
        """
        Description:
            This function retrieves the joint limits of the robot.
        """
        # Setup
        joint_indicies = self.plant.GetJointIndices(self.robot_model_idx)

        joint_limits = np.zeros((0, 2))
        for ii, joint_index in enumerate(joint_indicies):
            # Check to see if joint is actuated; if not, then ignore
            joint_ii = self.plant.get_joint(joint_index)
            joint_can_move = joint_ii.can_rotate() or joint_ii.can_translate()
            if not joint_can_move:
                continue

            lower_limits = self.plant.get_joint(joint_index).position_lower_limits()
            if len(lower_limits) == 0:
                lower_limits = [-np.inf]

            upper_limits = self.plant.get_joint(joint_index).position_upper_limits()
            if len(upper_limits) == 0:
                upper_limits = [np.inf]

            # Append limits to the joint_limits array
            joint_limits = np.vstack((
                joint_limits,
                [lower_limits[0], upper_limits[0]]
            ))

        return joint_limits

    def find_nearest_node(
        self,
        rrt: nx.DiGraph,
        q_random: np.ndarray
    ) -> Tuple[nx.classes.reportviews.NodeView, int]:
        """
        Description:
            This function finds the nearest node in the RRT to the random configuration.
        """
        nearest_node = None
        min_distance = float('inf')

        for node in rrt.nodes:
            distance = np.linalg.norm(rrt.nodes[node]['q'] - q_random)
            if distance < min_distance:
                min_distance = distance
                nearest_node_idx = node
                nearest_node = rrt.nodes[node]

        return nearest_node, nearest_node_idx

    def plan(
        self,
        q_start: np.ndarray,
        q_goal: np.ndarray,
        collision_check_fcn: Callable[[np.ndarray], bool] = None,
    ) -> Tuple[nx.DiGraph, bool]:
        """
        Description:
            This function executes the RRT planning algorithm.
        """
        # Input Processing
        q_start = q_start.flatten()
        q_goal = q_goal.flatten()
        if q_start.shape[0] != self.dim_q or q_goal.shape[0] != self.dim_q:
            raise ValueError(
                f"Start configuration shape ({q_start.shape}) and goal configuration " +
                f"shape ({q_goal.shape}) must match the dimension of the robot ({self.dim_q})."
            )

        if collision_check_fcn is None:
            collision_check_fcn =  self.check_collision_in_config

        # Setup
        rrt = nx.DiGraph()
        rrt.add_node(0, q=q_start) # Make this graph contain all configurations
        prob_sample_goal = self.config.prob_sample_goal

        # RRT Planner
        for iteration in range(self.config.max_iterations):
            # Sample random configuration OR goal
            if np.random.rand() < prob_sample_goal:
                q_random = q_goal.copy()
            else:
                q_random = self.sample_random_configuration()

            # Find nearest node in the tree
            nearest_node, nearest_node_idx = self.find_nearest_node(rrt, q_random)

            # Steer from nearest node to random configuration
            q_new = self.steer(nearest_node['q'], q_random)

            # Check for collisions
            if collision_check_fcn(q_new):
                continue

            # Add new node to the tree
            rrt.add_node(rrt.number_of_nodes(), q=q_new)
            rrt.add_edge(nearest_node_idx, rrt.number_of_nodes()-1)

            # Check if we have reached the goal
            if np.linalg.norm(q_new - q_goal) < self.config.convergence_threshold:
                print(f"Goal reached after {iteration} iterations!")
                return rrt, True

        # If we exit the loop without finding a path to the goal,
        # return the RRT and indicate failure
        print("Max iterations reached without finding a path to the goal.")
        return rrt, False



    def sample_random_configuration(self) -> np.ndarray:
        """
        Description:
            This function samples a random configuration within the joint limits.
        """
        return np.random.uniform(
            self.joint_limits[:, 0],
            self.joint_limits[:, 1]
        )

    def steer(
        self,
        nearest_node,
        q_random: np.ndarray
    ) -> np.ndarray:
        """
        Description:
            This function steers from the nearest node towards the random configuration.
        """
        # Setup
        step_size = self.config.steering_step_size
        q_nearest = nearest_node

        # Calculate the direction vector
        direction = q_random - q_nearest
        distance = np.linalg.norm(direction)

        # Either:
        # 1. Move a full step towards the random configuration
        # 2. Or, if the distance is less than the step size, return the random configuration
        if distance < step_size:
            return q_random

        return q_nearest + step_size * (direction / distance)



"""
cartesian_controller.py
Description:

    This file defines the CartesianController class. This class is used to control different manipulators and generates
    torque based control commands for the targeted arm.
"""
# External Imports
from pydrake.common.value import AbstractValue
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.inverse_kinematics import (
    DifferentialInverseKinematicsParameters,
    DoDifferentialInverseKinematics, InverseKinematics
)
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import MultibodyForces, JacobianWrtVariable, ModelInstanceIndex
from pydrake.solvers import Solve
from pydrake.systems.framework import BasicVector

import numpy as np

# Internal Imports
from brom_drake.control.arms.end_effector_target import EndEffectorTarget
from .base_arm_controller import BaseArmController


class CartesianArmController(BaseArmController):
    """
    A controller which imitates the cartesian control mode of the Kinova gen3 arm.

                                -----------------------------
                                |                           |
                                |                           |
    ee_target ----------------> |  CartesianArmController   | ----> applied_arm_torque
    ee_target_type -----------> |                           |
                                |                           |
                                |                           | ----> measured_ee_pose
    arm_joint_position -------> |                           | ----> measured_ee_twist
    arm_joint_velocity ------>  |                           |
                                |                           |
                                |                           |
                                -----------------------------

    Note that many ports are defined by the base class (BaseArmController).
    """

    def __init__(
        self,
        plant: MultibodyPlant,
        arm_model: ModelInstanceIndex,
        end_effector_frame_name: str = "end_effector_frame",
    ):
        BaseArmController.__init__(
            self,
            plant,
            arm_model,
            end_effector_frame_name=end_effector_frame_name,
        )

        # Define input and Output Ports
        self.ee_target_port, self.ee_target_type_port = None, None
        self.define_cartesian_target_input_ports()

        # Define output ports
        self.DeclareVectorOutputPort(
            "applied_arm_torque",
            BasicVector(self.plant.num_actuators()),
            self.CalcArmTorques)

        # Store desired end-effector pose and corresponding joint
        # angles so we only run full IK when we need to
        self.last_ee_pose_target = None
        self.last_q_target = None

    def define_cartesian_target_input_ports(self):
        """
        Description
        -----------
        Define the input ports for the CartesianArmController system
        :return:
        """
        self.ee_target_port = self.DeclareVectorInputPort(
            "ee_target",
            BasicVector(7),
        )
        self.ee_target_type_port = self.DeclareAbstractInputPort(
            "ee_target_type",
            AbstractValue.Make(EndEffectorTarget.kPose),
        )

    def CalcArmTorques(self, context, output):
        """
        This method is called each timestep to determine output torques
        """
        q = self.arm_joint_position_port.Eval(context)
        qd = self.arm_joint_velocity_port.Eval(context)
        self.plant.SetPositions(self.context, self.arm, q)
        self.plant.SetVelocities(self.context, self.arm, qd)

        # Some dynamics computations
        tau_g = -self.plant.CalcGravityGeneralizedForces(self.context)

        # Indicate what type of command we're recieving
        target_type = self.ee_target_type_port.Eval(context)

        if target_type == EndEffectorTarget.kWrench:
            # Compute joint torques consistent with the desired wrench
            wrench_des = self.ee_target_port.Eval(context)

            # Compute end-effector jacobian
            J = self.plant.CalcJacobianSpatialVelocity(self.context,
                                                       JacobianWrtVariable.kV,
                                                       self.ee_frame,
                                                       np.zeros(3),
                                                       self.world_frame,
                                                       self.world_frame)

            tau = tau_g + J.T @ wrench_des

            print("Calculated wrench torques")

        elif target_type == EndEffectorTarget.kTwist:
            # Compue joint torques consistent with the desired twist
            twist_des = self.ee_target_port.Eval(context)

            # Use DoDifferentialInverseKinematics to determine desired qd
            params = DifferentialInverseKinematicsParameters(self.plant.num_positions(),
                                                             self.plant.num_velocities())
            params.set_time_step(0.005)
            params.set_joint_velocity_limits((self.qd_min, self.qd_max))
            params.set_joint_position_limits((self.q_min, self.q_max))

            result = DoDifferentialInverseKinematics(self.plant,
                                                     self.context,
                                                     twist_des,
                                                     self.ee_frame,
                                                     params)

            if result.status == DifferentialInverseKinematicsStatus.kSolutionFound:
                qd_nom = result.joint_velocities
            else:
                print("Differential inverse kinematics failed!")
                qd_nom = np.zeros(self.plant.num_velocities())

            # Select desired accelerations using a proportional controller
            Kp = 10 * np.eye(self.plant.num_velocities())
            qdd_nom = Kp @ (qd_nom - qd)

            # Compute joint torques consistent with these desired accelerations
            f_ext = MultibodyForces(self.plant)
            tau = tau_g + self.plant.CalcInverseDynamics(self.context, qdd_nom, f_ext)

        elif target_type == EndEffectorTarget.kPose:
            # Compute joint torques which move the end effector to the desired pose
            rpy_xyz_des = self.ee_target_port.Eval(context)

            # Only do a full inverse kinematics solve if the target pose has changed
            if (rpy_xyz_des != self.last_ee_pose_target).any():

                X_WE_des = RigidTransform(
                    RollPitchYaw(rpy_xyz_des[:3]),
                    rpy_xyz_des[-3:],
                )

                # First compute joint angles consistent with the desired pose using Drake's IK.
                # This sets up a nonconvex optimization problem to find joint angles consistent
                # with the given constraints
                ik = InverseKinematics(self.plant, self.context)
                ik.AddPositionConstraint(self.ee_frame,
                                         [0, 0, 0],
                                         self.world_frame,
                                         X_WE_des.translation(),
                                         X_WE_des.translation())
                ik.AddOrientationConstraint(self.ee_frame,
                                            RotationMatrix(),
                                            self.world_frame,
                                            X_WE_des.rotation(),
                                            0.001)

                prog = ik.get_mutable_prog()
                q_var = ik.q()
                prog.AddQuadraticErrorCost(np.eye(len(q_var)), q, q_var)
                prog.SetInitialGuess(q_var, q)
                result = Solve(ik.prog())

                if not result.is_success():
                    print("Inverse Kinematics Failed!")
                    q_nom = np.zeros(self.plant.num_positions())
                else:
                    q_nom = result.GetSolution(q_var)

                    # Save the results of this solve for later
                    self.last_ee_pose_target = rpy_xyz_des
                    self.last_q_target = q_nom

            else:
                q_nom = self.last_q_target

            qd_nom = np.zeros(self.plant.num_velocities())

            # Use PD controller to map desired q, qd to desired qdd
            Kp = 1 * np.eye(self.plant.num_positions())
            Kd = 2 * np.sqrt(Kp)  # critical damping
            qdd_nom = Kp @ (q_nom - q) + Kd @ (qd_nom - qd)

            # Compute joint torques consistent with these desired qdd
            f_ext = MultibodyForces(self.plant)
            tau = tau_g + self.plant.CalcInverseDynamics(self.context, qdd_nom, f_ext)

        else:
            raise RuntimeError("Invalid target type %s" % target_type)
            print("Invalid target type")

        output.SetFromVector(tau)

        print("Calculated arm torques")


from brom_drake.DiagramTarget import DiagramTarget
from brom_drake.DiagramWatcher import DiagramWatcher
from .add_watcher import add_watcher, add_watcher_and_build, parse_list_of_simplified_targets
from brom_drake.PortWatcher.port_watcher import PortWatcher
from brom_drake.PortWatcher.port_watcher_options import PortWatcherOptions, FigureNamingConvention
from brom_drake.PortWatcher.port_figure_arrangement import PortFigureArrangement
from brom_drake.example_helpers import BlockHandlerSystem
from brom_drake.file_manipulation.urdf import DrakeReadyURDFConverter, drakeify_my_urdf

__all__ = [
    'DiagramTarget', 'DiagramWatcher',
    'add_watcher', 'add_watcher_and_build', 'parse_list_of_simplified_targets',
    'PortWatcher',
    'PortFigureArrangement', 'PortWatcherOptions', 'FigureNamingConvention',
    'BlockHandlerSystem',
    'DrakeReadyURDFConverter', 'drakeify_my_urdf',
]


"""
PortWatcherOptions.py
Description:

    This class defines the options used when making a PortWatcher object.
"""

from enum import Enum, IntEnum
from typing import NamedTuple

from .port_figure_arrangement import PortFigureArrangement

class FigureNamingConvention(IntEnum):
    kFlat = 0 # e.g. "plant_generalized_output_dim_0.png"
    kHierarchical = 1 # e.g. "system_plant/port_generalized_output/dim_0.png"

class PortWatcherOptions(NamedTuple):
    plot_arrangement: PortFigureArrangement = PortFigureArrangement.OnePlotPerPort
    plot_dpi: int = 300
    save_to_file: bool = True
    file_format: str = "png"
    figure_naming_convention: FigureNamingConvention = FigureNamingConvention.kFlat



from importlib import resources as impresources
import numpy as np
from pydrake.geometry import SceneGraph, Meshcat, MeshcatVisualizer
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import FixedOffsetFrame
from pydrake.systems.framework import Diagram, DiagramBuilder
from pydrake.systems.primitives import Demultiplexer

from . import EndEffectorWrenchCalculator
# Local imports
from .gripper_type import GripperType
from ..control.grippers import GripperController
from ..control.arms import (
    JointArmController, CartesianArmController,
    ArmControlMode,
)
from brom_drake.file_manipulation.urdf import DrakeReadyURDFConverter

from brom_drake import robots

class UR10eStation(Diagram):
    """
    A template system diagram for controlling a UR10e robot in a simulated environment.
    # TODO(kwesi): Draw the diagram for the two different control modes.
    """

    def __init__(
        self,
        time_step: float = 0.002,
        gripper_type: GripperType = GripperType.NoGripper,
        control_mode: ArmControlMode = ArmControlMode.kEndEffector,
    ):
        """
        Description:

            This class defines the
        """
        # Input Processing

        # Initialize the Diagram
        Diagram.__init__(self)
        self.set_name("UR10e_Station")

        # Create a diagram builder
        self.builder = DiagramBuilder()

        # Create scene_graph and plant
        self.scene_graph = self.builder.AddSystem(SceneGraph())
        self.scene_graph.set_name("UR10e_Station_SceneGraph")

        self.plant = self.builder.AddSystem(
            MultibodyPlant(time_step=time_step)
        )
        self.plant.RegisterAsSourceForSceneGraph(self.scene_graph)
        self.plant.set_name("UR10e_Station_Plant")

        # Add a shadow plant that only has access to the robot arm + gripper mass,
        # and is used for the controller. (Doesn't know about anything else in the scene)
        self.controller_plant = MultibodyPlant(time_step=time_step)
        self.controller_plant.set_name("UR10e_Station_ControllerPlant")

        # Body ID's and Poses for Anything else in the scene
        self.object_ids = []
        self.object_poses = []

        # Whether we have a camera in the simulation
        # self.has_camera = False

        self.arm_control_mode = control_mode
        self.AddArm()

        # Which sort of gripper we're using (if any)
        self.gripper_type = gripper_type
        if gripper_type == GripperType.Robotiq_2f_85:
            self.Add2f85Gripper()


    def AddArm(self):
        """
        Add the UR10e arm to the system.
        """
        # Setup
        arm_urdf_path = str(
            impresources.files(robots) / "models/ur/ur10e.urdf",
        )
        arm_urdf = DrakeReadyURDFConverter(arm_urdf_path).convert_urdf()
        arm_urdf = str(arm_urdf)

        self.end_effector_frame_name = "tool0"

        # The hardware system has lots of damping so this is more realistic,
        # but requires a simulation with small timesteps.

        self.arm = Parser(plant=self.plant).AddModels(arm_urdf)[0]
        self.controller_arm = Parser(plant=self.controller_plant).AddModels(arm_urdf)[0]

        # Fix the base of the arm to the world
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("base_link", self.arm),
        )

        self.controller_plant.WeldFrames(
            self.controller_plant.world_frame(),
            self.controller_plant.GetFrameByName("base_link", self.controller_arm),
        )

        # Create a new frame with the actual end-effector position.
        self.X_ee = RigidTransform()
        self.X_ee.set_translation([0, 0, 0.13])
        self.plant.AddFrame(FixedOffsetFrame(
            "end_effector",
            self.plant.GetFrameByName("tool0"),
            self.X_ee, self.arm))
        self.controller_plant.AddFrame(FixedOffsetFrame(
            "end_effector",
            self.controller_plant.GetFrameByName("tool0"),
            self.X_ee, self.controller_arm))

    def Add2f85Gripper(self):
        """
        Add the Robotiq 2F-85 gripper to the system. The arm must be added first.
        """
        # Setup

        # Add a gripper with actuation to the full simulated plant
        gripper_urdf_path = str(
            impresources.files(robots) / "models/2f_85_gripper/urdf/robotiq_2f_85.urdf"
        )
        self.gripper = Parser(plant=self.plant).AddModels(gripper_urdf_path)[0]

        X_grip = RigidTransform()
        X_grip.set_rotation(
            RotationMatrix(RollPitchYaw([0.0, 0.0, np.pi / 2]))
        )
        self.plant.WeldFrames(
            self.plant.GetFrameByName("tool0", self.arm),
            self.plant.GetFrameByName("robotiq_arg2f_base_link", self.gripper),
            X_grip,
        )

        # Add a gripper without actuation to the controller plant
        gripper_static_urdf = str(
            impresources.files(robots) / "models/2f_85_gripper/urdf/robotiq_2f_85_static.urdf"
        )
        static_gripper = Parser(plant=self.controller_plant).AddModels(
            gripper_static_urdf
        )[0]

        self.controller_plant.WeldFrames(
            self.controller_plant.GetFrameByName("tool0", self.controller_arm),
            self.controller_plant.GetFrameByName("robotiq_arg2f_base_link", static_gripper),
            X_grip)

    def ConnectToMeshcatVisualizer(self, port=None):
        self.meshcat = Meshcat(port)
        m = MeshcatVisualizer(self.meshcat)
        m.AddToBuilder(self.builder, self.scene_graph, self.meshcat)

        print("Open %s in a browser to view the meshcat visualizer." % self.meshcat.web_url())

    def Finalize(self):
        """
        Description
        ----------
        This finalizes the diagram by:
        - Finalizing all plants in the diagram
        :return: Nothing
        """
        # Setup

        # Finalize all plants
        self.plant.Finalize()
        self.controller_plant.Finalize()

        # Set up the scene graph
        self.builder.Connect(
            self.scene_graph.get_query_output_port(),
            self.plant.get_geometry_query_input_port()
        )

        self.builder.Connect(
            self.plant.get_geometry_pose_output_port(),
            self.scene_graph.get_source_pose_port(self.plant.get_source_id())
        )

        # Create Arm and Gripper Controllers
        self.CreateArmControllerAndConnect(self.arm_control_mode)

        if self.gripper_type != GripperType.NoGripper:
            self.CreateGripperControllerAndConnect()

        # Build the diagram
        self.builder.BuildInto(self)

    def CreateArmControllerAndConnect(
        self,
        arm_control_mode: ArmControlMode,
    ):
        """
        Description
        -----------
        This function creates a Cartesian arm controller and connects it to the rest of the system.
        :return:
        """

        controller = None

        if arm_control_mode == ArmControlMode.kJoint:
            controller = self.CreateJointArmControllerAndExportInputs()

        elif arm_control_mode == ArmControlMode.kEndEffector:
            controller = self.CreateCartesianArmControllerAndExportInputs()

        else:
            raise ValueError("Invalid Arm Control Mode: {}".format(arm_control_mode))

        # Output measured arm position and velocity
        demux = self.builder.AddSystem(
            Demultiplexer(self.plant.num_multibody_states(self.arm), self.plant.num_positions(self.arm)),
        )
        demux.set_name("UR10e_Station_Demux")

        self.builder.Connect(
            self.plant.get_state_output_port(self.arm),
            demux.get_input_port(0),
        )
        self.builder.ExportOutput(
            demux.get_output_port(0),
            "measured_arm_position",
        )
        self.builder.ExportOutput(
            demux.get_output_port(1),
            "measured_arm_velocity",
        )

        # Measured Arm Position and Velocity are sent to the controller
        self.builder.Connect(
            demux.get_output_port(0),
            controller.arm_joint_position_port,
        )
        self.builder.Connect(
            demux.get_output_port(1),
            controller.arm_joint_velocity_port,
        )

        # Torques from controller go to the "true" plant
        self.builder.Connect(
            controller.GetOutputPort("applied_arm_torque"),
            self.plant.get_actuation_input_port(self.arm),
        )

        # Controller outputs measured arm torques, end-effector pose, end-effector twist
        controller_output_port_names = ["applied_arm_torque", "measured_ee_pose", "measured_ee_twist"]
        for output_port_name in controller_output_port_names:
            self.builder.ExportOutput(
                controller.GetOutputPort(output_port_name),
                output_port_name,
            )

        # Compute and output end-effector wrenches based on measured joint torques
        wrench_calculator = self.builder.AddSystem(
            EndEffectorWrenchCalculator(
                self.controller_plant,
                self.controller_plant.GetFrameByName("end_effector"),
            ),
        )
        wrench_calculator.set_name("wrench_calculator")

        self.builder.Connect(
            demux.get_output_port(0),
            wrench_calculator.GetInputPort("joint_positions"))
        self.builder.Connect(
            demux.get_output_port(1),
            wrench_calculator.GetInputPort("joint_velocities"))
        self.builder.Connect(
            controller.GetOutputPort("applied_arm_torque"),
            wrench_calculator.GetInputPort("joint_torques"))

        self.builder.ExportOutput(
            wrench_calculator.get_output_port(),
            "measured_ee_wrench")

    def CreateCartesianArmControllerAndExportInputs(self) -> CartesianArmController:
        """
        Description
        -----------
        Defines the input ports for the CartesianArmController system

        :return:
        """
        # Create controller that uses the shadow plant to control the robot
        cartesian_controller = self.builder.AddSystem(
            CartesianArmController(
                self.controller_plant,
                self.controller_arm,
                end_effector_frame_name=self.end_effector_frame_name,
            ),
        )
        cartesian_controller.set_name("UR10e_Station_CartesianController")

        # End effector target and target type go to the controller
        self.builder.ExportInput(
            cartesian_controller.ee_target_port,
            "ee_target",
        )
        self.builder.ExportInput(
            cartesian_controller.ee_target_type_port,
            "ee_target_type",
        )

        return cartesian_controller


    def CreateJointArmControllerAndExportInputs(self)-> JointArmController:
        """
        Description
        -----------
        :return:
        """
        # Setup

        # Create Controller
        joint_controller = self.builder.AddSystem(
            JointArmController(
                self.controller_plant,
                self.controller_arm,
                end_effector_frame_name=self.end_effector_frame_name,
            ),
        )
        joint_controller.set_name("UR10e_Station_JointController")

        # End effector target and target type go to the controller
        self.builder.ExportInput(
            joint_controller.joint_target_port,
            "joint_target",
        )
        self.builder.ExportInput(
            joint_controller.joint_target_type_port,
            "joint_target_type",
        )

        return joint_controller

    def CreateGripperControllerAndConnect(self):
        """
        Description
        -----------
        This function creates a gripper controller and connects it to the rest of the system.
        Requires that the builder has been created and IS NOT FINALIZED.
        :return: (Nothing)
        """
        # Setup

        # Create gripper controller
        gripper_controller = self.builder.AddSystem(
            GripperController(self.gripper_type)
        )
        gripper_controller.set_name("gripper_controller")

        # Connect gripper controller to the diagram
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target"),
            "gripper_target")
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target_type"),
            "gripper_target_type")

        self.builder.Connect(
            self.plant.get_state_output_port(self.gripper),
            gripper_controller.GetInputPort("gripper_state"))
        self.builder.Connect(
            gripper_controller.GetOutputPort("applied_gripper_torque"),
            self.plant.get_actuation_input_port(self.gripper))

        # Send gripper position and velocity as an output
        self.builder.ExportOutput(
            gripper_controller.GetOutputPort("measured_gripper_position"),
            "measured_gripper_position")
        self.builder.ExportOutput(
            gripper_controller.GetOutputPort("measured_gripper_velocity"),
            "measured_gripper_velocity")


"""
Description
-----------
This file contains the URDF converter for transforming
arbitrary URDFs into URDFs that the Drake toolbox can
support.
"""
import os
from copy import copy, deepcopy
from typing import Tuple

import trimesh
from doctest import UnexpectedException

import loguru
from multiprocessing.managers import Value
from pathlib import Path
import xml.etree.ElementTree as ET

from .mesh_file_converter import MeshFileConverter
# Internal Supports
from .util import (
    does_drake_parser_support,
    URDF_CONVERSION_LOG_LEVEL_NAME, URDF_CONVERSION_LEVEL,
    tree_contains_transmission_for_joint,
    create_transmission_element_for_joint,
)


class DrakeReadyURDFConverter:
    def __init__(
        self,
        original_urdf_filename: str,
        output_urdf_file_path: str = None,
        overwrite_old_models: bool = False,
        overwrite_old_logs: bool = False,
        log_file_name: str = "conversion.log",
    ):
        # Setup
        self.original_urdf_filename = original_urdf_filename
        self.output_urdf_file_path = output_urdf_file_path
        self.log_file_name = log_file_name

        # Define conversion data directory in brom
        # (If it already exists, then we will overwrite it with the proper flag set)
        self.models_dir = Path("./brom/models/")
        if overwrite_old_logs and os.path.exists(self.models_dir / self.log_file_name):
            os.remove(self.models_dir / self.log_file_name)
        if overwrite_old_models:
            self.clean_up_models_dir()
        os.makedirs(self.models_dir, exist_ok=True)

        # Input Processing
        original_urdf_path = Path(self.original_urdf_filename)
        assert original_urdf_path.exists(), f"URDF file {original_urdf_path} does not exist!"

        if not (".urdf" in original_urdf_path.name):
            raise ValueError("Original URDF file must have a .urdf extension!")

        # Set Up logger
        self.configure_logger()
        DrakeReadyURDFConverter.log(
            f"Created a new DrakeReadyURDFConverter for file \"{self.original_urdf_filename}\"."
        )

        # Keep Track of the Joints In The Diagram
        self.actuated_joint_names = []

    def configure_logger(self):
        """
        Description
        -----------
        This method configures the loguru logger for the URDF conversion.
        :return:
        """
        # Setup
        urdf_conversion_level_exists = False
        log_file_name = self.log_file_name

        # Check to see if the log level exists
        try:
            found_level = loguru.logger.level(URDF_CONVERSION_LOG_LEVEL_NAME)
            urdf_conversion_level_exists = (found_level.name == URDF_CONVERSION_LOG_LEVEL_NAME)
        except Exception as e:
            # If the logger doesn't exist, then we will create it.
            expected_message = f"Level '{URDF_CONVERSION_LOG_LEVEL_NAME}' does not exist"
            if expected_message in str(e):
                pass
            else:
                raise UnexpectedException(f"Unexpected exception: {e}")

        # Remove the old logger
        loguru.logger.remove()

        # Configure logger if it doesn't exist
        if not urdf_conversion_level_exists:
            loguru.logger.level(
                URDF_CONVERSION_LOG_LEVEL_NAME,
                no=URDF_CONVERSION_LEVEL,
            )

        # Add logger
        loguru.logger.add(
            self.output_file_directory() / log_file_name,
            level=URDF_CONVERSION_LOG_LEVEL_NAME,
            filter=lambda record: record['level'].name == URDF_CONVERSION_LOG_LEVEL_NAME,
        )

    def convert_urdf(self) -> Path:
        """
        Description
        -----------
        Converts the original URDF file into a Drake-compatible URDF file.
        """
        # Setup
        original_urdf_path = Path(self.original_urdf_filename)
        original_xml = ET.ElementTree(file=original_urdf_path)

        # Use recursive function to convert the tree
        new_tree = self.convert_tree(original_xml)

        # Add transmissions, if needed
        for joint_name in self.actuated_joint_names:
            if tree_contains_transmission_for_joint(new_tree, joint_name):
                continue
            else:
                # If transmission doesn't exist in URDF, then add it!
                transmission_element = create_transmission_element_for_joint(joint_name)
                ET.indent(transmission_element, space="\t", level=0)
                new_tree.getroot().append(transmission_element)

        # Output the new tree to a file
        output_urdf_path = self.output_file_directory() / self.output_urdf_file_name()
        os.makedirs(output_urdf_path.parent, exist_ok=True)

        new_tree.write(output_urdf_path)

        DrakeReadyURDFConverter.log(
            f"Converted URDF file to Drake-compatible URDF file at {output_urdf_path}.\n\n"
        )
        return output_urdf_path

    def clean_up_models_dir(self):
        """
        Description
        -----------
        This method will clean up the models directory.
        :return:
        """
        # Make sure that the models directory exists
        os.makedirs(self.models_dir, exist_ok=True)
        os.system(f"rm -r {self.models_dir}")

    def convert_mesh_element(self, mesh_elt_in: ET.Element) -> ET.Element:
        # Setup
        new_elt = deepcopy(mesh_elt_in)

        # Algorithm
        self.log(
            f"Found a mesh element with filename \"{mesh_elt_in.attrib['filename']}\"."
        )

        # Check the value of the filename
        if not ("filename" in new_elt.attrib):
            raise ValueError(
                f"Found a mesh element that does not contain the \"filename\" attribute.\n"
                + "Brom doesn't know how to handle this!"
            )

        # Filename exists; Let's check to see if it's obj or not
        if does_drake_parser_support(new_elt.attrib["filename"]):
            pass
        else:
            # If parser does not support the given filename,
            # then let's try to create one
            old_filename = new_elt.attrib["filename"]
            new_filename = self.create_obj_to_replace_mesh_file(old_filename)
            new_elt.set(
                "filename", new_filename,
            )
            self.log(
                f"Replaced the mesh file \"{old_filename}\" with a Drake-compatible .obj file at \"{new_filename}\"."
            )

        return new_elt

    def convert_tree(
        self,
        current_tree: ET.ElementTree
    ) -> ET.ElementTree:
        """
        Description
        -----------`
        This recursive method is used to parse each element of the
        xml element tree.
        :param current_tree:
        :return:
        """
        # Setup
        root = deepcopy(current_tree.getroot())

        # Modify Root if it contains a mesh or file input
        new_root = self.convert_tree_element(root)

        # Remove all children from the new root
        for child in new_root.findall("*"):
            new_root.remove(child)

        # Modify Children
        for child in root:
            new_child = self.convert_tree(
                ET.ElementTree(child)
            )
            new_root.append(new_child.getroot())

        return ET.ElementTree(new_root)


    def convert_tree_element(
        self,
        elt: ET.Element,
    ) -> ET.Element:
        """
        Description
        -----------
        This method is used to transform one element in the XML tree into an eleemnt
        that that Drake URDF parser can handle. If necessary, it will create a new
        3d model file and place it into the right directory.
        :param elt: An element in the ElementTree that we would like to convert
                     into a Drake-ready element.
        :return: The Drake-ready xml tree element.
        """
        # Setup
        new_elt = deepcopy(elt)

        # Convert file paths for mesh elements if necessary
        if new_elt.tag == "mesh":
            new_elt = self.convert_mesh_element(new_elt)
        elif new_elt.tag == "transmission":
            # Ignore transmission elements during the conversion
            return new_elt
        elif new_elt.tag == "joint":
            self.handle_joint_element(new_elt)
        else:
            # If the element is not a mesh element, then we will just return a copy of it.
            self.log(
                f"Found an element with tag \"{elt.tag}\" that is not a mesh element. No modification needed..."
            )
        # Otherwise, just return a copy of the previous element.
        return new_elt

    def create_obj_to_replace_mesh_file(
        self,
        mesh_file_name: str,
    ) -> str:
        """
        Description
        -----------
        This function will create an .obj file to replace the
        .stl or .dae file that is given in mesh file "mesh_file_name".
        :param mesh_file_name:
        :return:
        """
        # Setup
        original_urdf_dir = Path(self.original_urdf_filename).parent

        # Convert the file
        converter = MeshFileConverter(
            mesh_file_path=mesh_file_name,
            urdf_dir=Path(original_urdf_dir),
            new_urdf_dir=self.output_file_directory(),
        )
        new_mesh_path = converter.convert()

        # Clip off all parts of path that include the exported output
        new_mesh_path = str(new_mesh_path).replace(str(self.output_file_directory()), "")

        return "./" + str(new_mesh_path)


    def handle_joint_element(self, joint_elt: ET.Element):
        """
        Description
        -----------
        This method will handle the joint element in the URDF file.
        :param joint_elt:
        :return:
        """
        # Setup

        # Keep track of all FREE joint names

        # If the joint does not have a "type" attribute, then we will assume it is fixed
        # and won't add it.
        if "type" not in joint_elt.attrib:
            return

        # If joint is "fixed", then don't add it to the list of actuated joints
        joint_is_fixed = joint_elt.attrib["type"] == "fixed"
        if joint_is_fixed:
            return

        joint_name = joint_elt.attrib["name"]
        self.actuated_joint_names.append(joint_name)

    @staticmethod
    def log(message: str):
        """
        Description
        -----------
        Logs a message to the logger.
        :param message: A string with the message we want to send to the logs.
        :return:
        """
        loguru.logger.log(URDF_CONVERSION_LOG_LEVEL_NAME, message)

    def output_file_directory(self) -> Path:
        """
        Description
        -----------
        Returns the directory of the output urdf file.
        :return:
        """
        # Setup
        original_file_path = Path(self.original_urdf_filename)

        # Input Processing
        if self.output_urdf_file_path is None:
            return self.models_dir / original_file_path.name.replace(".urdf", "")
        else:
            output_file_path = Path(self.output_urdf_file_path)
            return output_file_path.parent

    def output_urdf_file_name(self) -> str:
        """
        Description
        -----------
        Creates the output filename based on the original filename.
        """
        # Setup
        original_file_path = Path(self.original_urdf_filename)

        if self.output_urdf_file_path is None:
            # Use Pathlib to extract the filename, if it exists
            original_name = original_file_path.name
            original_name = original_name.replace(".urdf", "")

            return f"{original_name}.drake.urdf"
        else:
            return Path(self.output_urdf_file_path).name


import numpy as np
from pydrake.math import RollPitchYaw
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import JacobianWrtVariable, ModelInstanceIndex
from pydrake.systems.framework import LeafSystem, BasicVector


class BaseArmController(LeafSystem):
    """
    A controller which provides the basic functionality for controlling
    a robotic arm.
                                ------------------------
                                |                      |
                                |                      |
    ee_target ----------------> |  BaseArmController   | ----> applied_arm_torque
    ee_target_type -----------> |                      |
                                |                      |
                                |                      | ----> measured_ee_pose
    arm_joint_position -------> |                      | ----> measured_ee_twist
    arm_joint_velocity ------>  |                      |
                                |                      |
                                |                      |
                                ------------------------
    The type of target is determined by ee_target_type, and the options are defined in the
    end effector target.
    """
    def __init__(
        self,
        plant: MultibodyPlant,
        arm_model: ModelInstanceIndex,
        end_effector_frame_name: str = "end_effector_frame",
    ):
        LeafSystem.__init__(self)

        self.plant = plant
        self.arm = arm_model
        self.context = self.plant.CreateDefaultContext()

        # Define some relevant frames
        self.world_frame = self.plant.world_frame()
        self.ee_frame = self.plant.GetFrameByName(end_effector_frame_name)

        # Define the base input and output ports
        self.arm_joint_position_port, self.arm_joint_velocity_port = None, None
        self.define_input_ports_for_arm_state()

        self.define_output_measurement_ports()

        # Set joint limits (set self.{q,qd}_{min,max})
        self.q_min, self.q_max = None, None
        self.qd_min, self.qd_max = None, None
        self.GetJointLimits()

    def define_input_ports_for_arm_state(self):
        """
        Description
        -----------
        Defines the input ports for the arm state. This is useful for
        connecting the arm state to the controller.
        :return:
        """
        self.arm_joint_position_port = self.DeclareVectorInputPort(
            "arm_joint_position",
            BasicVector(self.plant.num_positions(self.arm)),
        )
        self.arm_joint_velocity_port = self.DeclareVectorInputPort(
            "arm_joint_velocity",
            BasicVector(self.plant.num_velocities(self.arm)),
        )

    def define_output_measurement_ports(self):
        """
        Description
        -----------
        Defines the output ports for the controller that have to deal with
        the sensed state of the input arm.
        :return:
        """
        self.DeclareVectorOutputPort(
            "measured_ee_pose",
            BasicVector(6),
            self.CalcEndEffectorPose,
            {self.time_ticket()}  # indicate that this doesn't depend on any inputs,
        )  # but should still be updated each timestep
        self.DeclareVectorOutputPort(
            "measured_ee_twist",
            BasicVector(6),
            self.CalcEndEffectorTwist,
            {self.time_ticket()})

    def CalcEndEffectorPose(self, context, output):
        """
        This method is called each timestep to determine the end-effector pose
        """
        q = self.arm_joint_position_port.Eval(context)
        qd = self.arm_joint_velocity_port.Eval(context)
        self.plant.SetPositions(self.context, self.arm, q)
        self.plant.SetVelocities(self.context, self.arm, qd)

        # Compute the rigid transform between the world and end-effector frames
        X_ee = self.plant.CalcRelativeTransform(
            self.context,
            self.world_frame,
            self.ee_frame,
        )

        ee_pose = np.hstack([
            RollPitchYaw(X_ee.rotation()).vector(),
            X_ee.translation()
        ])

        output.SetFromVector(ee_pose)

    def CalcEndEffectorTwist(self, context, output):
        """
        This method is called each timestep to determine the end-effector twist
        """
        q = self.arm_joint_position_port.Eval(context)
        qd = self.arm_joint_velocity_port.Eval(context)
        self.plant.SetPositions(self.context, self.arm, q)
        self.plant.SetVelocities(self.context, self.arm, qd)

        # Compute end-effector Jacobian
        J = self.plant.CalcJacobianSpatialVelocity(
            self.context,
            JacobianWrtVariable.kV,
            self.ee_frame,
            np.zeros(3),
            self.world_frame,
            self.world_frame,
        )

        ee_twist = J @ qd
        output.SetFromVector(ee_twist)

        # print("Calculated end effector twist")

    def GetJointLimits(self):
        """
        Iterate through self.plant to establish joint angle
        and velocity limits.

        Sets:

            self.q_min
            self.q_max
            self.qd_min
            self.qd_max

        """
        q_min = []
        q_max = []
        qd_min = []
        qd_max = []

        joint_indices = self.plant.GetJointIndices(self.arm)

        for idx in joint_indices:
            joint = self.plant.get_joint(idx)

            if joint.type_name() == "revolute":  # ignore the joint welded to the world
                q_min.append(joint.position_lower_limit())
                q_max.append(joint.position_upper_limit())
                qd_min.append(joint.velocity_lower_limit())  # note that higher limits
                qd_max.append(joint.velocity_upper_limit())  # are availible in cartesian mode

        self.q_min = np.array(q_min)
        self.q_max = np.array(q_max)
        self.qd_min = np.array(qd_min)
        self.qd_max = np.array(qd_max)

        # print("Got joint limits")


from typing import Union, Tuple, List

from pydrake.systems.framework import DiagramBuilder, LeafSystem, Diagram, Context

from brom_drake.all import add_watcher_and_build
# Internal Imports
from brom_drake.scenes.roles import Role
from brom_drake.scenes.ids import SceneID
from brom_drake.utils import Performer

class BaseScene:
    """
    Base class for all scenes.
    """
    def __init__(self, **kwargs):
        # Create a builder for the scene
        self.builder = DiagramBuilder()

        # Create the diagram and diagram_context placeholders
        self.diagram = None
        self.diagram_context = None

        # Save a list of the performers
        self.performers = []

    def add_all_secondary_cast_members_to_builder(self):
        """
        Description
        -----------
        This method must be implemented by the subclass. It should add all
        elements to the builder.
        :return:
        """
        pass

    def suggested_roles(self) -> List[Role]:
        """
        Description
        -----------
        This method should be implemented by the subclass.
        It should return a list of roles that are suggested for the scene.
        :return:
        """
        return []

    def fill_role(
        self,
        role: Role,
        system: Performer,
    ):
        """
        Description
        -----------
        This method should be implemented by the subclass. It should add the
        system to the role.
        :param role:
        :param system:
        :return:
        """
        # Setup
        builder = self.builder

        # Call the member method of the role object
        role.connect_performer_ports_to(builder, system)

        # Save the performer
        self.performers.append(system)

    def cast_scene(
        self,
        cast: Tuple[Role, Performer] = [],
    ):
        # Setup

        # Add all elements to the builder
        self.add_all_secondary_cast_members_to_builder()

        # Fulfill each role-performer pair in the casting_call list
        for role, performer in cast:
            self.fill_role(role, performer)

    def cast_scene_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
    ) -> Tuple[Diagram, Context]:
        # Setup
        builder = self.builder
        self.cast_scene(cast)

        # Build the diagram
        # self.diagram = builder.Build()
        # self.diagram_context = self.diagram.CreateDefaultContext()

        watcher, self.diagram, self.diagram_context = add_watcher_and_build(builder)

        return self.diagram, self.diagram_context

    @property
    def id(self) -> SceneID:
        """
        Description
        -----------
        The unique identifier for the scene
        :return:
        """
        return SceneID.kNotDefined


"""
PortWatcher.py
Description:

    This file defines the PortWatcher class. This class is used to watch the ports of a
    diagram.
"""
from pathlib import Path
from typing import List, Tuple, Union, NamedTuple
import loguru
import numpy as np
import matplotlib.pyplot as plt
import os

from pydrake.systems.framework import OutputPort, PortDataType, DiagramBuilder, LeafSystem
from pydrake.systems.primitives import LogVectorOutput
from pydrake.systems.framework import Context

# Internal Imports
from brom_drake.DiagramWatcher.constants import DEFAULT_PLOT_DIR
from .port_watcher_options import PortWatcherOptions, FigureNamingConvention
from .port_figure_arrangement import PortFigureArrangement


class PortWatcher:
    def __init__(
        self,
        system: LeafSystem,
        output_port: OutputPort,
        builder: DiagramBuilder,
        logger_name: str = None,
        options: PortWatcherOptions = PortWatcherOptions(),
        plot_dir: str = DEFAULT_PLOT_DIR,
    ):
        # Setup
        self.options = options
        self.system = system
        self.port = output_port
        self.data = {}
        self.plot_handles = {}
        self.plot_handles = None
        self.plot_dir = plot_dir

        # Input Processing
        if output_port.get_data_type() != PortDataType.kVectorValued:
            raise ValueError(
                f"This watcher only supports vector valued ports (i.e., of type {PortDataType.kVectorValued}.\n" +
                f"Received port of type {output_port.get_data_type()}."
            )

        if logger_name is None:
            logger_name = f"PortWatcher_{system.get_name()}_{output_port.get_name()}"

        # Processing
        self.logger = LogVectorOutput(output_port, builder)
        self.logger.set_name(logger_name)

    def plot_logger_data(
        self,
        diagram_context: Context,
    ) -> Tuple[List[plt.Figure], List[List[plt.Axes]]]:
        """
        plot_logger_data
        Description:

            This function plots the data in the logger.
        :param diagram_context:
        :return:
        """
        # Setup

        # Get the log from the logger
        temp_log = self.logger.FindLog(diagram_context)

        log_times = temp_log.sample_times()
        log_data = temp_log.data()

        if (log_data.shape[1] == 0) or (log_data.shape[0] == 0):
            loguru.logger.warning(
                f"No data found for {self.system.get_name()} - Port {self.port.get_name()}! Skipping...")
            return None, None

        # Plot the data
        if self.options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            fig, ax_list = self.plot_logger_data_subplots(diagram_context, log_times, log_data)
            return [fig], [ax_list]

        elif self.options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            figs, ax_grid = [], []
            for port_index in range(self.port.size()):
                fig_ii = plt.figure()
                ax_ii = fig_ii.add_subplot(1, 1, 1)
                ax_ii.plot(
                    log_times, log_data[port_index, :]
                )

                # Save figures and axes to lists
                figs.append(fig_ii)
                ax_grid.append([ax_ii])

            return figs, ax_grid

        else:
            raise ValueError(
                f"Invalid plot arrangement: {self.options.plot_arrangement}."
            )

    def plot_logger_data_subplots(
        self,
        diagram_context: Context,
        times: np.array,
        data: np.array,

    ):
        """
        plot_logger_data_subplots
        Description:

            This function plots the data in the logger.
        :param times:
        :param data:
        :param diagram_context:
        :return:
        """
        # Setup
        n_dims = data.shape[0]

        # Plot the data
        n_rows, n_cols = self.compute_plot_shape(n_dims)

        print(f"Plotting {n_dims} dimensions in a {n_rows}x{n_cols} grid.")

        fig, ax_list = plt.subplots(n_rows, n_cols)

        if n_rows == 1 and n_cols == 1:
            ax_list.plot(times, data[0, :])
            ax_list.set_title(f"Dim #0")

        else:
            for row_index in range(n_rows):
                for col_index in range(n_cols):

                    dim_index = n_cols * row_index + col_index

                    if dim_index >= n_dims:
                        fig.delaxes(ax_list[row_index, col_index])
                        continue

                    ax_list[row_index, col_index].plot(times, data[dim_index, :])
                    ax_list[row_index, col_index].set_title(f"Dim #{dim_index}")

        return fig, ax_list

    def compute_plot_shape(self, n_dims: int) -> Tuple[int, int]:
        """
        Description:
            Computes the shape of the plot based on the data.
        :param n_dims: The number of dimensions in the data.
        :return:
        """
        if n_dims == 1:
            return 1, 1
        if n_dims == 2:
            return 1, 2

        if n_dims < 9:
            return 2, int(np.ceil(n_dims / 2.0))

        # Otherwise
        return 3, int(np.ceil(n_dims / 3.0))

    def safe_system_name(self) -> str:
        """
        Description:
            Returns a safe name for the system.
        :param name: System's name.
        :return:
        """
        out = self.system.get_name()
        # First, let's check to see how many "/" exist in the name
        slash_occurences = [i for i, letter in enumerate(out) if letter == "/"]
        if len(slash_occurences) > 0:
            out = out[slash_occurences[-1] + 1:]  # truncrate string based on the last slash

        # Second, replace all spaces with underscores
        out = out.replace(" ", "_")

        return out

    def save_figures(self, diagram_context: Context):
        """
        savefigures
        Description:
            This function saves the figures.
        TODO(kwesi): Make it so that this function computes names + directory structure based on plot arrangement.
        :param diagram_context:
        :return:
        """
        # Setup
        figs, ax_list = self.plot_logger_data(diagram_context)

        # If no figures are returned, then return early!
        if figs is None:
            return

        if len(figs) == 0:
            return # Do nothing

        # Save the figures
        figure_names = self.figure_names()

        if len(figs) == 1:
            # Create directory for the plots, if it doesn't already exist
            os.makedirs(Path(figure_names[0]).parent, exist_ok=True)
            figs[0].savefig(figure_names[0], dpi=self.options.plot_dpi)
        else:
            # Plot each figure within this directory
            for ii, fig_ii in enumerate(figs):
                os.makedirs(Path(figure_names[ii]).parent, exist_ok=True)
                fig_ii.savefig(figure_names[ii],dpi=self.options.plot_dpi)

        # Close all figures when done
        plt.close('all')

    def figure_names(self) -> List[str]:
        """
        Description:
            Returns the name of the figure.
        :return:
        """
        # Setup
        options = self.options

        # If this has the flat naming convention, then the file should be contained within the plot_dir.
        if options.figure_naming_convention == FigureNamingConvention.kFlat:
            return self.figure_names_under_flat_convention()


        elif options.figure_naming_convention == FigureNamingConvention.kHierarchical:
            return self.figure_names_under_hierarchical_convention()

        else:
            raise NotImplementedError(
                f"Invalid figure naming convention for figure_names(): {options.figure_naming_convention}."
            )


    def figure_names_under_flat_convention(self) -> List[str]:
        """
        Description:
            Returns the names associated with each figure that this port will
            generate assuming we are under the kFlat convention.
        :param self:
        :return: List of strings where each string is a file name for an associated figure.
        """
        # Setup
        options = self.options
        format = options.file_format

        #
        if options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            return [
                f"{self.plot_dir}/system_{self.safe_system_name()}_port_{self.port.get_name()}.{format}"
            ]

        elif options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            return [
                f"{self.plot_dir}/system_{self.safe_system_name()}_port_{self.port.get_name()}_dim{ii}.{format}"
                for ii in range(self.port.size())
            ]

        else:
            raise NotImplementedError(
                f"Invalid plot arrangement for figure naming convention {options.figure_naming_convention}: {options.plot_arrangement}."
            )

    def figure_names_under_hierarchical_convention(self) -> List[str]:
        """
        Description:
            Returns the names associated with each figure that this port will
            generate assuming we are under the kHierarchical convention.
        :return:
        """
        # Setup
        options = self.options
        format = options.file_format

        # Algorithm
        if options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            return [
                f"{self.plot_dir}/system_{self.safe_system_name()}/port_{self.port.get_name()}.{format}"
            ]

        elif options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            return [
                f"{self.plot_dir}/system_{self.safe_system_name()}/port_{self.port.get_name()}/dim{ii}.{format}"
                for ii in range(self.port.size())
            ]

        else:
            raise NotImplementedError(
                f"Invalid plot arrangement for figure naming convention {options.figure_naming_convention}: {options.plot_arrangement}."
            )


import loguru
import os
from pathlib import Path
from typing import Tuple
import xml.etree.ElementTree as ET

import trimesh

# Internal imports
from brom_drake.file_manipulation.urdf.util import URDF_CONVERSION_LOG_LEVEL_NAME


class MeshFileConverter:
    def __init__(
        self,
        mesh_file_path: str,
        urdf_dir: Path,
        new_urdf_dir: Path,
    ):
        """
        Description:
            This class will convert a mesh file to a different format.
        :param brom_models_dir:
        :param mesh_file:
        """
        # Input Processing

        # Setup
        self.mesh_file = mesh_file_path
        self.urdf_dir = urdf_dir
        self.new_urdf_dir = new_urdf_dir

    def convert(
        self,
        output_file_path: Path = None,
    ):
        # Setup
        if output_file_path is None:
            output_file_path = self.define_output_path(output_file_path)

        # Use trimesh to convert
        # print(f"urdf_dir: {self.urdf_dir}")
        # print(f"true_path: {self.true_mesh_file_path()}")
        # print(
        #     str(self.urdf_dir / self.true_mesh_file_path())
        # )
        mesh = trimesh.load_mesh(
            str(self.urdf_dir / self.true_mesh_file_path())
        )

        os.makedirs(self.define_output_path(output_file_path).parent, exist_ok=True)
        mesh.export(self.define_output_path(output_file_path))

        # Define the relative output path
        return self.define_output_path(output_file_path)

    @property
    def output_mesh_directory(self) -> Path:
        new_urdf_dir = self.new_urdf_dir
        return new_urdf_dir

    def define_output_path(self, output_mesh_file: Path = None) -> Path:
        """
        Description:
            This function will define the output path for the mesh file.
        :param output_mesh_file:
        :return:
        """
        # Input Processing
        if output_mesh_file is not None:
            return output_mesh_file

        # Setup
        mesh_file_name = self.mesh_file

        # Check to see if the file type is supported
        for suffix in self.supported_suffixes:
            if suffix in str(mesh_file_name):
                # Strip off all parts of the file that might contain:
                # - The package name
                # - "../"
                # - "./"
                clean_mesh_file_name = str(mesh_file_name)
                if "package://" in clean_mesh_file_name:
                    clean_mesh_file_name = clean_mesh_file_name.replace("package://", "")
                    clean_mesh_file_name = clean_mesh_file_name[
                        clean_mesh_file_name.find("/") + 1:
                    ]
                elif "file://" in clean_mesh_file_name:
                    clean_mesh_file_name = clean_mesh_file_name.replace("file://", "")
                    if os.path.isabs(clean_mesh_file_name):
                        # If path is absolute, then let's save the new mesh file in a simplified way.
                        clean_mesh_file_name = Path(clean_mesh_file_name).name
                    else:
                        clean_mesh_file_name = clean_mesh_file_name[
                            clean_mesh_file_name.find("/") + 1:
                        ]


                clean_mesh_file_name = clean_mesh_file_name.replace("../", "")
                clean_mesh_file_name = clean_mesh_file_name.replace("./", "")

                # If we pass this point, then we know that the file type is supported
                # Clean all parts of the file name that are not the name
                output_mesh_file = clean_mesh_file_name.replace(suffix, ".obj")

                # Prepend a models directory to the path
                return self.output_mesh_directory / output_mesh_file

        # If we pass this point, then we know that the file type is not supported
        raise ValueError(
            f"File type {mesh_file_name} is not supported by this function!\n" +
            f"Make sure that the file type is one of the following: {self.supported_suffixes}"
        )

    def find_file_path_in_package(
        self,
        max_depth: int = 10,
    ) -> Path:
        """
        Description
        -----------
        This method will find the file path in the package.

        :param max_depth: The maximum depth to search for the package directory.
        :return:
        """

        # Setup
        mesh_file_name = self.mesh_file

        # Extract the package name and the package directory from mesh file name
        package_dir, package_name = self.find_package_directory_including_mesh(max_depth=max_depth)

        # Create the output file path from what we now know
        return package_dir / Path(mesh_file_name.replace(f"package://{package_name}/", ""))

    def find_package_directory_including_mesh(
        self,
        max_depth: int = 10,
    ) -> Tuple[Path, str]:
        # Setup
        original_urdf_dir = Path(self.urdf_dir)

        # Find the path to the package
        package_found = False
        candidate_path = original_urdf_dir
        search_depth = 1
        while not package_found:
            print(f"candidate_path: {candidate_path}")
            # Check to see if "package.xml" exists in the directory
            if (candidate_path / "package.xml").exists():
                self.log(f"Found package path at {candidate_path}.")
                package_found = True
            else:
                candidate_path = candidate_path.parent
                search_depth += 1

            if search_depth > max_depth:
                raise ValueError(
                    f"Could not find package path for file {original_urdf_dir} after searching {max_depth} directories."
                )

        # Should now have found the package directory
        package_dir = candidate_path

        # Open the package.xml file and find the name of the package
        package_xml = ET.ElementTree(file=package_dir / "package.xml")
        package_root = package_xml.getroot()
        package_name = package_root.find("name").text

        return package_dir, package_name

    @staticmethod
    def log(message: str):
        """
        Description
        -----------
        Logs a message to the logger.
        :param message: A string with the message we want to send to the logs.
        :return:
        """
        loguru.logger.log(URDF_CONVERSION_LOG_LEVEL_NAME, message)

    def mesh_file_path_is_relative(self):
        """
        Description:
            This function will check if the mesh file path is relative.
        """
        # Algorithm
        mesh_file = self.mesh_file

        # Check to see if file path starts with "./"
        if mesh_file.startswith("./"):
            return True

        # Ignore this if the mesh file contains a package prefix
        if mesh_file.startswith("package:"):
            return False

        # Check to see if the first part of the path contains a folder or file
        # in the current directory
        if mesh_file.startswith("file://"):
            mesh_file = mesh_file.replace("file://", "")
            return not os.path.isabs(mesh_file)

        # Now, check to see if the target file exists
        complete_file_path = self.urdf_dir / Path(mesh_file)
        exists = complete_file_path.exists()
        return exists

        # raise FileNotFoundError(
        #     f"File {complete_file_path} does not exist in directory {os.getcwd()}!"
        # )

    @property
    def supported_suffixes(self):
        return [".stl", ".dae", ".DAE", ".STL"]

    def true_mesh_file_path(self, max_depth: int = 10) -> Path:
        """
        Description:
            This function will return the true mesh file path.
        """
        # Algorithm
        if self.mesh_file.startswith("package:"):
            return self.find_file_path_in_package(max_depth=max_depth)
        elif self.mesh_file_path_is_relative():
            mesh_file = self.mesh_file.replace("file://", "")
            return Path(mesh_file)
        elif os.path.isabs(self.mesh_file.replace("file://", "")):
            return Path(self.mesh_file.replace("file://", ""))
        else:
            raise ValueError(
                f"Mesh file path {self.mesh_file} is not supported by this function!\n" +
                "Make sure that the \"filename\" attribute contains a relative path (e.g., starts with \"./\")\n"
                " or a package path (i.e., starts with \"package:\")."
            )


"""
Description:
    This file will contain utilities for quickly creating simple URDF files
    to be used in debugging and testing.
"""
# External Imports
from dataclasses import dataclass
from enum import IntEnum
import os
import xml.etree.ElementTree as ET

import numpy as np

# Internal Imports
from .shape_definition import ShapeEnum, SphereDefinition, ShapeDefinition

@dataclass
class InertiaDefinition:
    """
    A dataclass that defines the inertia of a simple shape.
    """
    ixx: float = 1.0
    ixy: float = 0.0
    ixz: float = 0.0
    iyy: float = 1.0
    iyz: float = 0.0
    izz: float = 1.0

    def as_list(self) -> list:
        """
        Convert the inertia to a list.
        :return: A list of the inertia values.
        """
        return [self.ixx, self.ixy, self.ixz, self.iyy, self.iyz, self.izz]

    def as_map(self) -> dict:
        """
        Convert the inertia to a map.
        :return: A map of the inertia values.
        """
        return {
            "ixx": f"{self.ixx}",
            "ixy": f"{self.ixy}",
            "ixz": f"{self.ixz}",
            "iyy": f"{self.iyy}",
            "iyz": f"{self.iyz}",
            "izz": f"{self.izz}",
        }

@dataclass
class SimpleShapeURDFDefinition:
    """
    A dataclass that defines a simple shape URDF.
    """
    name: str
    shape: ShapeDefinition
    color: np.ndarray = None
    create_collision: bool = True # Whether to create the collision elements of the urdf
    mass: float = 1.0
    inertia: InertiaDefinition = None

    def as_urdf(self) -> ET.Element:
        """
        Create the URDF for the simple shape using python's built-in xml library.
        :return:
        """
        # Setup
        root = ET.Element("robot", {"name": self.name + "_robot"})
        link = ET.SubElement(root, "link", {"name": self.name + "_base_link"})

        # Add inertial elements to link
        self.add_inertial_elements_to(link)

        # Add visual elements to link
        self.add_visual_elements_to(link)

        # Add collision elements to link
        if self.create_collision:
            self.add_collision_elements_to(link)

        return root

    def add_inertial_elements_to(self, link_elt: ET.Element):
        """
        Add the inertial elements to the link element.
        :param link_elt: The ET.Element object for the link.
        :return: Nothing, but modifies the link_elt in place.
        """
        # Setup
        inertial_link = ET.SubElement(link_elt, "inertial")
        inertia_def = self.inertia
        if inertia_def is None:
            inertia_def = InertiaDefinition()

        # Create all child links for inertial elements
        origin_elt = ET.SubElement(
            inertial_link,
            "origin",
            {"xyz": "0 0 0", "rpy": "0 0 0"},
        )
        mass = ET.SubElement(inertial_link, "mass", {"value": f"{self.mass}"})
        inertia = ET.SubElement(
            inertial_link,
            "inertia",
            inertia_def.as_map(),
        )

        # Add inertial elements to link
        link_elt.append(inertial_link)

    def add_visual_elements_to(self, link_elt: ET.Element):
        """
        Add the visual elements to the link element.
        :param link_elt: The ET.Element object for the link.
        :return: Nothing, but modifies the link_elt in place.
        """
        # Setup
        visual_link = ET.SubElement(link_elt, "visual")
        color = self.color
        if color is None:
            color = np.array([0.0, 1.0, 0.0, 0.8])

        # Create all child links for visual elements
        origin_elt = ET.SubElement(
            visual_link,
            "origin",
            {"xyz": "0 0 0", "rpy": "0 0 0"},
        )
        geometry = ET.SubElement(visual_link, "geometry")

        # Add the shape to the geometry element
        if self.shape.type == ShapeEnum.kSphere:
            sphere = ET.SubElement(geometry, "sphere", {"radius": f"{self.shape.radius}"})
        else:
            raise ValueError(f"Shape {self.shape} not supported yet.")

        # Add the material to the visual element
        material = ET.SubElement(visual_link, "material", {"name": "blue"})
        color = ET.SubElement(
            material,
            "color",
            {"rgba": f"{color[0]} {color[1]} {color[2]} {color[3]}"},
        )

        # Add visual elements to link
        link_elt.append(visual_link)

    def add_collision_elements_to(self, link_elt: ET.Element):
        """
        Add the collision elements to the link element.
        :param link_elt: The ET.Element object for the link.
        :return:
        """
        # Setup
        collision_link = ET.SubElement(link_elt, "collision")
        origin_elt = ET.SubElement(
            collision_link,
            "origin",
            {"xyz": "0 0 0", "rpy": "0 0 0"},
        )
        geometry = ET.SubElement(collision_link, "geometry")

        # Add the shape to the geometry element
        if self.shape.type == ShapeEnum.kSphere:
            sphere = ET.SubElement(geometry, "sphere", {"radius": f"{self.shape.radius}"})
        else:
            raise ValueError(f"Shape {self.shape} not supported yet.")

        # Add collision elements to link
        link_elt.append(collision_link)

    def write_to_file(self, file_path: str):
        """
        Write the URDF to a file.
        :param file_path: The path to the file where the URDF will be written.
        :return: Nothing, but writes the URDF to the file.
        """
        # Setup
        root = self.as_urdf()
        tree = ET.ElementTree(root)
        ET.indent(tree, space="\t", level=0)

        # Write to file
        tree.write(file_path, xml_declaration=True)






import xml.etree.ElementTree

URDF_CONVERSION_LOG_LEVEL_NAME = "BROM_URDF_CONVERSION"
URDF_CONVERSION_LEVEL = 21

def does_drake_parser_support(filename: str):
    """
    Description
    -----------
    This function cleanly answers whether the given 3d object
    file is supported by Drake.
    :param filename:
    :return:
    """
    return ".obj" in filename # TODO(kwesi): Determine if .sdf should be put here.

def create_transmission_element_for_joint(
    actuated_joint_name: str,
) -> xml.etree.ElementTree.Element:
    """
    Description
    -----------
    This function creates a transmission element for the given joint.
    :param actuated_joint_name: The name of the joint that will be actuated
    :return: XML Element defining the new transmission for the actuated joint
    """

    # Create the transmission element
    transmission = xml.etree.ElementTree.Element("transmission")

    # Create an inner type element
    transmission_type_element = xml.etree.ElementTree.Element("type")
    transmission_type_element.text = "transmission_interface/SimpleTransmission"

    # Create inner joint element (reference to the actuated_joint_name
    joint_element = xml.etree.ElementTree.Element("joint")
    joint_element.set("name", actuated_joint_name)

    # Create inner actuator element
    actuator_element = xml.etree.ElementTree.Element("actuator")
    actuator_element.set("name", f"{actuated_joint_name}_actuator")

    # Assemble transmission element
    transmission.append(transmission_type_element)
    transmission.append(joint_element)
    transmission.append(actuator_element)

    return transmission

def tree_contains_transmission_for_joint(
    tree: xml.etree.ElementTree.ElementTree,
    actuated_joint_name: str,
) -> bool:
    """
    Description
    -----------
    This function determines if the given tree contains a transmission
    element for the given actuated joint.
    :param tree:
    :param actuated_joint_name:
    :return:
    """
    # Setup
    root = tree.getroot()

    # Check to see if the root is a transmission element
    if root.tag == "transmission":
        # Check if the transmission element contains the actuated joint
        for child in root:
            if child.tag == "joint" and child.attrib["name"] == actuated_joint_name:
                return True

    # Check each of the tree's children to see if there is a transmission element
    for child in root:
        child_tree_contains_transmission_for_joint = tree_contains_transmission_for_joint(
            xml.etree.ElementTree.ElementTree(child),
            actuated_joint_name,
        )
        if child_tree_contains_transmission_for_joint:
            return True

    # If we've searched through the full sub-tree and don't see the actuated joint,
    # then we return False
    return False


"""
Description:
    This file defines an OfflineMotionPlanner class that is used to define the
    offline motion planning role that can be used in Brom scenes.
"""
from brom_drake.scenes.roles import Role, RolePortAssignment
from brom_drake.scenes.roles.role_port_assignment import PairingType

kOfflineMotionPlanner = Role(
    name="OfflineMotionPlanner",
    description="This role is used to define the offline motion planning role.",
    port_assignments=[
        RolePortAssignment(
            "start_pose", "start_pose",
            pairing_type=PairingType.kInput,
        ),
        RolePortAssignment(
            "goal_pose", "goal_pose",
            pairing_type=PairingType.kInput,
        ),
        RolePortAssignment(
            external_target_name="robot_model_index_source",
            performer_port_name="robot_model_index",
            pairing_type=PairingType.kInput,
        ),
        RolePortAssignment(
            "id", "scene_id",
            is_required=False,
            pairing_type=PairingType.kInput,
        ),
        RolePortAssignment(
            external_target_name="plan",
            performer_port_name="motion_plan",
            pairing_type=PairingType.kOutput,
        ),
        RolePortAssignment(
            external_target_name="plan_ready",
            performer_port_name="plan_is_ready",
            is_required=False,
            pairing_type=PairingType.kOutput,
        ),
    ],
)



"""
motion_planner.py
Description:

        This file defines the MotionPlanner class.
        This class is used to plan motion for a robot.
"""
import numpy as np
from pydrake.geometry import SceneGraph
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex


class MotionPlanner:
    def __init__(
        self,
        robot_model_idx: ModelInstanceIndex,
        plant: MultibodyPlant,
        scene_graph: SceneGraph,
    ):
        """
        Description:
            This function initializes the motion planner.
        """
        # Input Processing
        self.robot_model_idx = robot_model_idx
        self.plant = plant
        self.scene_graph = scene_graph

        # Setup
        self.root_context = None # Usually, the diagram context that we can use to extract the plant's context

    def check_collision_in_config(
        self,
        q_model: np.ndarray,
    ) -> bool:
        """
        Description:
            This function checks for collisions in the robot's environment.
        """
        # Input Processing
        if self.root_context is None:
            raise ValueError("Plant context is not initialized yet!")

        # Setup
        plant_context = self.plant.GetMyMutableContextFromRoot(self.root_context)
        scene_graph_context = self.scene_graph.GetMyMutableContextFromRoot(self.root_context)

        # Set the configuration
        self.plant.SetPositions(
            plant_context,
            self.robot_model_idx,
            q_model
        )

        # Note: This method will be inaccurate if using arbitrary, nonconvex meshes!
        # https://drake.mit.edu/pydrake/pydrake.geometry.html?highlight=queryobject#pydrake.geometry.QueryObject.HasCollisions
        query_object = self.scene_graph.get_query_output_port().Eval(scene_graph_context)
        return query_object.HasCollisions()




    def plan_motion(self):
        """
        Description:
            This function plans the motion for the robot.
        """
        # Motion planning logic goes here
        pass


from brom_drake.control.arms.cartesian_arm_controller import CartesianArmController
from brom_drake.control.arms.end_effector_target import EndEffectorTarget
from .ideal_joint_position_controller import IdealJointPositionController

__all__ = [
    "CartesianArmController",
    "EndEffectorTarget",
    "IdealJointPositionController",
]


from importlib import resources as impresources
from typing import Tuple, Callable

import networkx as nx
import numpy as np
from pydrake.common.eigen_geometry import Quaternion
from pydrake.common.value import AbstractValue
from pydrake.geometry import GeometrySet, CollisionFilterDeclaration
from pydrake.math import RigidTransform, RollPitchYaw
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Diagram, Context
from pydrake.systems.primitives import ConstantValueSource

# Internal Imports
import brom_drake.robots as robots
from brom_drake.example_helpers import AddGround
from brom_drake.motion_planning.systems.open_loop_plan_dispenser import OpenLoopPlanDispenser
from brom_drake.robots.stations.kinematic import UR10eStation as KinematicUR10eStation
from brom_drake.scenes import SceneID
from brom_drake.scenes.roles import Role
from brom_drake.scenes.types.motion_planning import OfflineMotionPlanningScene
from brom_drake.utils import Performer


class ShelfPlanningScene(OfflineMotionPlanningScene):
    def __init__(
        self,
        time_step=1e-3,
        shelf_pose: RigidTransform = None, # The pose of the shelf
        meshcat_port_number: int = 7001, # Usually turn off for CI (i.e., make it None)
        plan_execution_speed: float = 0.2,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.time_step = time_step
        self.shelf_pose = shelf_pose
        self.meshcat_port_number = meshcat_port_number
        self.plan_execution_speed = plan_execution_speed

        # Input Processing
        if self.shelf_pose is None:
            shelf_orientation = Quaternion(1, 0, 0 , 0)
            shelf_position = np.array([-0.4, 0.8, 0.0])
            self.shelf_pose = RigidTransform(
                RollPitchYaw(0.0, 0.0, +np.pi/2.0).ToQuaternion(),
                np.array([0.0, 1.0, 0.6]),
            )

        self.desired_cupboard_positions = np.array([0.0, 0.0])

        # Create containers for meshcat and other values we will set later
        self.meshcat = None
        self.plan_dispenser = None

        # Set station
        self.station = KinematicUR10eStation(
            time_step=self.time_step,
            meshcat_port_number=self.meshcat_port_number,
        )
        self.arm = self.station.arm

        # Set Names of Plant and scene graph
        self.plant = self.station.plant
        self.scene_graph = self.station.scene_graph

        self.plant.set_name(f"ShelfScene_Plant")
        self.scene_graph.set_name(f"ShelfScene_SceneGraph")

        # Create variables for the models we have
        self.shelf_model_index = None
        self.geometry_ids_to_ignore = []

    def add_all_secondary_cast_members_to_builder(self):
        """
        Add all secondary cast members to the builder.
        :return:
        """
        # Setup
        self.add_ur10e_station() # Use the UR10e station's plant + scene graph for all other objects

        # Add obstacles
        self.add_shelf(self.plant)

        # Add the ground as well as the start and goal locations
        AddGround(self.plant)
        self.add_start_and_goal_to_this_plant(self.plant)

        self.station.Finalize()

        # Add The Motion Planning Components (e.g., the interpolator)
        self.add_start_source_system()
        self.add_goal_source_system()
        self.add_robot_source_system()

        self.add_motion_planning_components()

        # Connect motion planning components to station
        # self.connect_motion_planning_components()

        # Connect to Meshcat
        # if self.use_meshcat:
        #     self.meshcat = Meshcat(port=7001)  # Object provides an interface to Meshcat
        #     m_visualizer = MeshcatVisualizer(self.meshcat)
        #     m_visualizer.AddToBuilder(builder, self.scene_graph, self.meshcat)

        # Print message to user
        print("Added all secondary cast members to the builder.")

    def add_motion_planning_components(self):
        """
        Add the motion planning components to the builder.
        :return:
        """
        # Setup
        n_actuated_dof = self.plant.num_actuated_dofs()

        # Add the Plan Dispenser and connect it to the station
        self.plan_dispenser = self.builder.AddSystem(
            OpenLoopPlanDispenser(n_actuated_dof, self.plan_execution_speed)
        )

        self.builder.Connect(
            self.plan_dispenser.GetOutputPort("point_in_plan"),
            self.station.GetInputPort("desired_joint_positions"),
        )

    def add_robot_source_system(self):
        """
        Description
        -----------
        This method adds a source for providing the motion planner
        with the model index for the robot that we are trying to control.
        :return:
        """
        # Setup

        # Create AbstractValueSource
        robot_source_system = ConstantValueSource(
            AbstractValue.Make(self.station.arm)
        )
        robot_source_system.set_name("robot_model_index_source")

        self.builder.AddSystem(robot_source_system)

    def add_shelf(self, plant: MultibodyPlant):
        """
        Add the shelf to the scene.
        :return:
        """
        # Setup
        urdf_file_path = str(
            impresources.files(robots) / "models/cupboard/cupboard.sdf"
        )

        # Add the shelf to the plant
        model_idcs = Parser(plant=plant).AddModels(urdf_file_path)
        self.shelf_model_index = model_idcs[0]

        # Weld the bookshelf to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("cupboard_body", self.shelf_model_index),
            self.shelf_pose,
        )

    def add_ur10e_station(self):
        """
        Add the Kinematic UR10e station to the scene for simple motion planning.
        :return:
        """
        # Setup

        # Add station
        self.builder.AddSystem(self.station)

    def cast_scene_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        Modifies the normal cast_scene_and_build, so that
        we share the context of the plant with the appropriate
        parts of the system.
        :param cast:
        :return:
        """
        diagram, diagram_context = super().cast_scene_and_build(cast=cast)

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        for role_ii, performer_ii in cast:
            if role_ii.name == "OfflineMotionPlanner":
                performer_ii.set_internal_root_context(
                    diagram_context
                )

        return diagram, diagram_context

    def configure_collision_filter(self, scene_graph_context: Context):
        """
        Description
        -----------
        This method configures the collision filter for the scene.
        :param scene_graph_context:
        :return:
        """
        # Setup
        scene_graph = self.scene_graph
        shelf_model_index = self.shelf_model_index

        # Ignore self collisions of the shelf, by:
        # - Getting the model instance index of the shelf
        # - Collecting the Geometry IDs of all the geometries in the shelf
        geometry_ids = []
        for body_index in self.plant.GetBodyIndices(shelf_model_index):
            # Get the geometry IDs
            geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )
            print(self.plant.get_body(body_index))

        self.geometry_ids_to_ignore = geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.geometry_ids_to_ignore)
            )
        )

    def connect_motion_planning_components(
        self,
    ):
        """
        Description
        -----------
        This method connects the motion planning components directly to the
        kinematic controller.
        :return:
        """

        # Connect the plan dispenser to the station
        pass

    def easy_cast_and_build(
        self,
            planning_algorithm: Callable[
                [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
                Tuple[nx.DiGraph, np.ndarray],
            ],
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        This function is used to easily cast and build the scene.
        :param planning_algorithm: The algorithm that we will use to
        :return:
        """
        # Setup

        # Use Base class implementation to start
        diagram, diagram_context = super().easy_cast_and_build(planning_algorithm)

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        self.performers[0].set_internal_root_context(
            diagram_context
        )

        return diagram, diagram_context

    @property
    def goal_pose(self):
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        goal_position = np.array([+0.1, 1.0, 0.55]) # np.array([+0.5, 0.7, 0.65])
        goal_orientation = RollPitchYaw(np.pi/2.0, np.pi/2.0, 0.0).ToQuaternion()
        return RigidTransform(goal_orientation, goal_position)

    @property
    def id(self) -> SceneID:
        return SceneID.kShelfPlanning1

    @property
    def start_pose(self):
        """
        Get the start pose. This should be defined by the subclass.
        :return:
        """
        start_position = np.array([+0.3, 0.1, 1.2])
        start_orientation = Quaternion(1, 0, 0, 0)
        return RigidTransform(start_orientation, start_position)


from enum import IntEnum
import numpy as np
from pydrake.common.value import AbstractValue
from pydrake.systems.framework import LeafSystem, Context, BasicVector
from pydrake.trajectories import PiecewiseTrajectory, PiecewisePolynomial


class StateOfPlanInMemory(IntEnum):
    kNotSet = 0
    kSet = 1

class OpenLoopPlanDispenser(LeafSystem):
    def __init__(
        self,
        n_dof: int,
        speed: float, # Speed at which to proceed through points.
    ):
        LeafSystem.__init__(self)

        # Setup
        self.n_dof = n_dof
        self.speed = speed
        self.plan, self.planned_trajectory = None, None

        # Create input port for the plan (we assume that when we receive it, we will close the port)
        self.plan_is_set = False
        sample_plan = np.zeros((0, self.n_dof))
        self.t0, self.t_final = -1.0, -1.0
        self.plan_port = self.DeclareAbstractInputPort(
            "plan",
            AbstractValue.Make(sample_plan),
        )

        self.plan_ready_port = self.DeclareAbstractInputPort(
            "plan_ready",
            AbstractValue.Make(False),
        )

        # Create output ports
        self.DeclareVectorOutputPort(
            "point_in_plan",
            self.n_dof,
            self.GetCurrentPointInPlan,
        )

        self.state_of_plan_in_memory_idx = self.DeclareDiscreteState(
            np.array([StateOfPlanInMemory.kNotSet])
        )

        self.DeclareAbstractOutputPort(
            "plan_is_set",
            lambda: AbstractValue.Make(StateOfPlanInMemory.kNotSet),
            self.GetStateOfPlanInMemory,
            {
                self.discrete_state_ticket(
                    self.state_of_plan_in_memory_idx,
                )
            },  # This output should only be updated when the
                # discrete state state_of_plan_in_memory changes
        )

    def GetCurrentPointInPlan(self, context: Context, output_point: BasicVector):
        # Setup
        plan_is_ready = self.plan_ready_port.Eval(context)
        plan = self.plan_port.Eval(context)

        # Get the abstract state
        abstract_state = context.get_discrete_state(self.state_of_plan_in_memory_idx)

        # If plan isn't ready, then skip the rest of the logic
        if not plan_is_ready:
            output_point.SetFromVector(
                np.zeros((self.n_dof,))
            )
            return

        # If this is the first time that plan_is_ready, then
        # let's save the current time as the time when the plan starts
        if abstract_state[0] == StateOfPlanInMemory.kNotSet:
            # Initialize the system to handle the new plan
            self.initialize_system_for_new_plan(context)

        # Use interpolation to get the current point
        t = context.get_time() - self.t0

        # Mark the plan as set
        context.SetDiscreteState(np.array([
            StateOfPlanInMemory.kSet,
        ]))

        # Output The Current Point
        if t > self.t_final:
            output_point.SetFromVector(
                self.planned_trajectory.value(self.t_final).flatten()
            )
        else:
            output_point.SetFromVector(
                self.planned_trajectory.value(t).flatten()
            )




    def find_time_between_two_points(self, x1: np.ndarray, x2: np.ndarray)->float:
        """
        Description
        -----------
        This method computes the time that it will take to move between two points.
        :return:
        """
        # Setup

        # Compute the distance between the two points
        dist = np.linalg.norm(x1 - x2)
        return dist / self.speed

    def initialize_system_for_new_plan(self, context: Context):
        """
        Description
        -----------
        This function should lock a couple of things into memory for this object to handle future
        calls to get the plan.

        :param context:
        :return:
        """
        # Setup

        # Get time to start executing the plan
        self.t0 = context.get_time()
        context.SetDiscreteState(
            np.array([StateOfPlanInMemory.kSet])
        )

        # Get times where we should reach each point in the plan and save it into a new map
        self.plan = self.plan_port.Eval(context)

        n_points_in_plan = self.plan.shape[0]
        t_ii = 0.0
        times = [t_ii]
        for ii in range(n_points_in_plan-1):
            t_ii += self.find_time_between_two_points(self.plan[ii,:], self.plan[ii+1, :])
            times += [t_ii]

        self.t_final = t_ii
        self.planned_trajectory = PiecewisePolynomial.FirstOrderHold(times, self.plan.T) # Need transpose to make each column a sample





    def GetStateOfPlanInMemory(self, context: Context, output: AbstractValue):
        """Plan is set if and only if the internal variable is not None. """
        output.SetFrom(
            context.get_abstract_state(self.state_of_plan_in_memory_idx)[0]
        )




from .joint_target import JointTarget
from .arm_control_mode import ArmControlMode
from .base_arm_controller import BaseArmController
from .end_effector_target import EndEffectorTarget
from .joint_arm_controller import JointArmController
from .cartesian_arm_controller import CartesianArmController

__all__ = [
    "ArmControlMode",
    "JointTarget",
    "BaseArmController",
    "EndEffectorTarget",
    "JointArmController",
    "CartesianArmController",
]



from dataclasses import dataclass
from distutils.command.build import build
from enum import IntEnum
from platform import system
from typing import Union, List

from pydrake.systems.framework import DiagramBuilder, LeafSystem

from brom_drake.utils import (
    Performer, find_all_systems_with_output_port, find_all_systems_with_input_port,
)
from brom_drake.utils.search import find_all_systems_with_name

PortName = str
SystemName = str
TargetName = Union[PortName, SystemName]


class PairingType(IntEnum):
    kInput = 1
    kOutput = 2

@dataclass
class RolePortAssignment:
    external_target_name: TargetName  # The name of the port (or system) that we want to connect this port to
    performer_port_name: PortName
    pairing_type: PairingType
    is_required: bool = True  # By default, a port pairing is required

    def create_connections_in(
        self,
        builder: DiagramBuilder,
        performer: Performer,
    ):
        """
        Description
        -----------

        :param builder:
        :param performer:
        :return:
        """

        if self.pairing_type == PairingType.kInput:
            self.create_input_port_connections_in(
                builder=builder,
                performer=performer,
            )
        elif self.pairing_type == PairingType.kOutput:
            self.create_output_port_connections_in(
                builder=builder,
                performer=performer,
            )
        else:
            raise ValueError(
                f"Unknown pairing type: {self.pairing_type}"
            )

    def create_input_port_connections_in(
        self,
        builder: DiagramBuilder,
        performer: Performer,
    ):
        # Setup

        # Check to see if the performer has the given input port
        performer_has_input_port = performer.HasInputPort(
            self.performer_port_name
        )
        if (not performer_has_input_port) and self.is_required:
            raise self.create_assignment_port_unavailable_error()
        elif (not performer_has_input_port) and (not self.is_required):
            print(
                f"Performer {performer.get_name()} does not have OPTIONAL input port named {self.performer_port_name}." +
                "\nWill skip its connection."
            )
            return # Do not do anything if the port does not exist

        # Otherwise, port exists.
        performer_port = performer.GetInputPort(self.performer_port_name)

        # Get the external system that has the output port with the correct name
        systems_list = self.find_any_matching_output_targets(builder)
        assert len(systems_list) == 1, self.create_not_enough_systems_error(systems_list)

        # Connect system
        system_is_target = systems_list[0].get_name() == self.external_target_name
        if system_is_target:
            builder.Connect(
                systems_list[0].get_output_port(),
                performer_port,
            )
        else:
            builder.Connect(
                systems_list[0].GetOutputPort(self.external_target_name),
                performer_port,
            )

    def find_any_matching_output_targets(
        self,
        builder: DiagramBuilder,
    ) -> List[LeafSystem]:
        """
        Description
        -----------
        This function attempts to find all targets that match the output
        target definition, i.e.:
        - A system with the output port with given name
        - A system with the given name.
        :param builder:
        :return:
        """
        # Setup
        external_target_name = self.external_target_name

        # Retrieve all systems that have that output port name external_target_name
        systems_list = find_all_systems_with_output_port(
            builder, external_target_name,
        )

        if len(systems_list) > 0:
            return systems_list

        # Retrieve all systems that have the name of external_target_name
        systems_list = find_all_systems_with_name(builder, external_target_name)

        if len(systems_list) > 0:
            return systems_list

        # Otherwise, raise an error
        raise self.create_no_target_found_error()


    def create_output_port_connections_in(
        self,
        builder: DiagramBuilder,
        performer: Performer,
    ):
        """
        Description
        -----------
        Creates the connections from the performer's output port to
        the external system's input ports.
        :param builder:
        :param performer:
        :return:
        """
        # Setup

        # Check to see if the performer has the given input port
        performer_has_output_port = performer.HasOutputPort(
            self.performer_port_name
        )
        if (not performer_has_output_port) and self.is_required:
            raise self.create_assignment_port_unavailable_error()
        elif (not performer_has_output_port) and (not self.is_required):
            print(
                f"Performer {performer.get_name()} does not have OPTIONAL output port named {self.performer_port_name}." +
                "\nWill skip its connection."
            )
            return  # Do not do anything if the port does not exist

        # Otherwise, port exists.
        performer_port = performer.GetOutputPort(self.performer_port_name)

        # Get the external system that has the input port with the correct name
        systems_list = self.find_any_matching_input_targets(builder)
        assert len(systems_list) == 1, self.create_not_enough_systems_error(systems_list)

        # TODO(kwesi): Create if-else statement to handle the cases of when systems_list
        # contains either:
        # 1. A system with the name given by external_target_name
        # 2. A system that has a port with name given by external_target_name
        builder.Connect(
            performer_port,
            systems_list[0].GetInputPort(self.external_target_name),
        )

    def find_any_matching_input_targets(
        self,
        builder: DiagramBuilder,
    ) -> List[LeafSystem]:
        """
        Description
        -----------
        This function attempts to find all targets that match the input
        target definition, i.e.:
        - A system with the input port with given name
        - A system with the given name.
        :param builder:
        :return:
        """
        # Setup
        external_target_name = self.external_target_name

        # Retrieve all systems that have that output port name external_target_name
        systems_list = find_all_systems_with_input_port(
            builder, external_target_name,
        )

        if len(systems_list) > 0:
            return systems_list

        # Retrieve all systems that have the name of external_target_name
        systems_list = find_all_systems_with_name(builder, external_target_name)

        if len(systems_list) > 0:
            return systems_list

        # Otherwise, raise an error
        raise self.create_no_target_found_error()

    def create_assignment_port_unavailable_error(self) -> ValueError:
        # Setup
        port_type_str = "UNKNOWN"
        if self.pairing_type == PairingType.kInput:
            port_type_str = "INPUT"
        elif self.pairing_type == PairingType.kOutput:
            port_type_str = "OUTPUT"

        # Create the error message
        return ValueError(
            f"Performer does not have required {port_type_str} port \"{self.performer_port_name}\""
        )

    def create_not_enough_systems_error(self, systems_list: List[LeafSystem]) -> str:
        # Setup
        port_type_str = "UNKNOWN"
        if self.pairing_type == PairingType.kInput:
            port_type_str = "INPUT"
        elif self.pairing_type == PairingType.kOutput:
            port_type_str = "OUTPUT"

        # Create message depending on the pairing type
        return f"Expected 1 system to have {port_type_str} port \"{self.external_target_name}\"," + \
            f" but found {len(systems_list)} systems with that {port_type_str} port."

    def create_no_target_found_error(self):
        # Setup
        external_target_name = self.external_target_name

        # Determine if this is an input or output target
        external_target_type = "UNKNOWN"
        if self.pairing_type == PairingType.kInput:
            external_target_type = "OUTPUT"
        elif self.pairing_type == PairingType.kOutput:
            external_target_type = "INPUT"

        # Create error
        return ValueError(
            f"Failed to find an external system with the name {external_target_name}" +
            f" or an external system with {external_target_type} port name {external_target_name}." +
            f"\nCheck your RolePortAssignment definition."
        )


"""
Description:
    This file contains a set of helpful utilities that can be used to help with the
    development of the robot models.
"""

from typing import List
import xml.etree.ElementTree as ET

def find_base_link_name_in(path_to_robot_model: str) -> str:
    """
    Description
    -----------
    This method will try to find a good base link for the model.
    :param path_to_robot_model: The path to the robot model.
    :return:
    """
    # Setup
    path_to_model = path_to_robot_model

    # Parse the model using xml
    if ".urdf" in path_to_model:
        original_xml = ET.ElementTree(file=path_to_model)
        link_names = find_all_link_names(original_xml)

        # Find a link that contains the word "base"
        for link_name in link_names:
            if "base" in link_name.lower():
                return link_name # Return the first one we find

    else:
        raise ValueError(
            "We can only smartly find base links in .urdf files, for now.\n" +
            "File an issue if you want more support in the future."
        )

    # If we can't find a base link, then we will raise an error
    raise ValueError(
        "We could not find a good base link in the model.\n" +
        "Please provide the base link name manually by adding \"base\" to one of the urdf's links."
    )


def find_all_link_names(xml_tree: ET.ElementTree ) -> List[str]:
    """
    Description
    -----------
    This method will find all the link names in the xml tree.
    :param xml_tree: The xml tree that we would like to investigate.
    :return:
    """
    # Setup
    link_names = []

    # Find all the links
    for link in xml_tree.findall(".//link"):
        link_names.append(link.attrib["name"])

    return link_names


"""
DiagramWatcher.py
Description:

    Creates a DiagramWatcher class that can be used to monitor the state of a Diagram
    and automatically log signals of interest.
"""
import os
from typing import List, Union
import matplotlib.pyplot as plt
import numpy as np
import loguru

from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Diagram, DiagramBuilder, LeafSystem, PortDataType
from pydrake.systems.primitives import (
    VectorLogSink, ConstantVectorSource, AffineSystem, LogVectorOutput,
)

# Internal Imports
from brom_drake.DiagramTarget import DiagramTarget
from brom_drake.PortWatcher.port_watcher import PortWatcher
from brom_drake.PortWatcher.port_watcher_options import PortWatcherOptions
from brom_drake.PortWatcher.port_watcher import FigureNamingConvention
from brom_drake.DiagramWatcher.constants import INELIGIBLE_SYSTEM_TYPES, DEFAULT_PLOT_DIR
from brom_drake.DiagramWatcher.errors import UnrecognizedTargetError


class DiagramWatcher:
    def __init__(
        self,
        subject: DiagramBuilder,
        targets: List[DiagramTarget] = None,
        plot_dir: str = DEFAULT_PLOT_DIR,
        port_watcher_options: PortWatcherOptions = PortWatcherOptions(),
    ):
        # Setup

        # Needs to be populated by the user of this class AFTER the diagram has been built
        self.diagram = None
        self.diagram_context = None

        # Check subject
        if not isinstance(subject, DiagramBuilder):
            raise ValueError("subject must be a DiagramBuilder!")

        # Save the inputs
        self.subject = subject
        self.plot_dir = plot_dir

        # Create the .brom directory, to store:
        # - activity_summary.log
        # - all plots
        if os.path.exists(plot_dir):
            os.system(f"rm -r {plot_dir}")

        # Create directory to plot in
        os.makedirs(plot_dir, exist_ok=True)
        self.configure_brom_activity_summary()  # Create an "activity summary" log
                                                # which details what the
                                                # DiagramWatcher is doing.

        # Collect All the Connections and Systems
        self.eligible_systems = self.find_eligible_systems(subject)
        if targets is None:
            targets = [DiagramTarget(system.get_name()) for system in self.eligible_systems]
        else:
            self.check_targets(targets, self.eligible_systems)

        # Log the list of eligible systems
        loguru.logger.info(f"Found {len(self.eligible_systems)} systems in diagram are eligible for targeting:")
        for idx, system in enumerate(self.eligible_systems):
            loguru.logger.info(f"{idx}: {system.get_name()}")

        # For Each Target with None ports, we will try to
        # "smartly" create the targets that we want to monitor
        inferred_targets = self.get_smart_targets(subject, targets)
        self.inferred_targets = inferred_targets

        # For each target's port, we will add a logger
        self.port_watchers = {target.name: {} for target in inferred_targets}
        loguru.logger.info("Adding loggers to the diagram... (via PortWatcher objects)")
        for target in inferred_targets:
            system = subject.GetSubsystemByName(target.name)
            for port_index in target.ports:
                target_port = system.get_output_port(port_index)

                if target_port.get_data_type() != PortDataType.kVectorValued:
                    print(f"Port {port_index} of system {target.name} is not vector-valued! Skipping...")
                    loguru.logger.warning(
                        f"Port {port_index} ({target_port.get_name()}) of system {target.name} is not vector-valued! " +
                        "Will not add a logger for it.",
                    )
                    continue

                # Configure PortWatcher
                self.port_watchers[target.name][target_port.get_name()] = PortWatcher(
                    system, target_port, subject,
                    logger_name=f"{target.name}_logger_{port_index}",
                    plot_dir=plot_dir,
                    options=port_watcher_options,
                )

    def __del__(self):
        """
        Description:

            Destructor for the Diagram Watcher.
            Will plot the data from all of our loggers if we have access to the diagram context.
        :return:
        """
        # Setup

        is_ready_to_plot = self.diagram is not None

        # Upon deletion, we will PLOT the data from all of our loggers
        # if we have access to the diagram context
        if is_ready_to_plot:
            self.save_figures()


    def configure_brom_activity_summary(self):
        """
        Description:
            Configures the "activity summary" a log of brom's activity.
        :return:
        """
        # Setup
        loguru.logger.remove()  # Remove the default logger
        loguru.logger.add(self.plot_dir + "/activity_summary.log")

    def check_targets(
        self,
        targets: List[DiagramTarget],
        eligible_systems: List[Union[MultibodyPlant, AffineSystem, LeafSystem]],
    ) -> List[Union[MultibodyPlant, AffineSystem, LeafSystem]]:
        """
        Description:
            Finds the systems specified by the targets list that
            we want to watch/monitor.
            We will try to ignore all systems that are:
            - Scene Graphs
            - Loggers
            and raise an error if the target is not found in the eligible systems.
        :param targets:
        :param eligible_systems:
        :return:
        """

        # Find all the systems that are eligible for logging
        eligible_system_dict = {
            system.get_name(): system for system in eligible_systems
        }

        # Search for each target in eligible_systems
        targeted = []
        for target in targets:
            # Check if the target name is in the eligible systems
            if target.name not in eligible_system_dict.keys():
                raise UnrecognizedTargetError(target, eligible_system_dict.keys())

            # If it is, then also check that the port index is correct
            if target.ports is None:
                continue  # No need to check things if ports is None

            num_ports_in_target = eligible_system_dict[target.name].num_output_ports()
            for port_index in target.ports:
                if port_index < 0 or port_index >= num_ports_in_target:
                    raise ValueError(f"Port index {port_index} is out of bounds for system {target.name} (only {num_ports_in_target} ports exist)")

        # All checks passed!
        pass

    def find_eligible_systems(
        self,
        builder: DiagramBuilder
    ) -> List[Union[MultibodyPlant, AffineSystem, LeafSystem]]:
        """
        Description:
            Finds all the systems that are eligible for logging.
            We want to ignore all systems that are:
            - Scene Graphs
            - Loggers
        """

        # Find all the systems that are eligible for logging
        eligible_systems = []

        loguru.logger.info("Finding all eligible systems for logging...")
        for system in builder.GetSystems():
            if type(system) in INELIGIBLE_SYSTEM_TYPES:
                loguru.logger.warning(
                    f"System {system.get_name()} (of type {type(system)}) is not eligible for logging! Skipping..."
                )
                continue

            # Otherwise add to list
            eligible_systems.append(system)

        return eligible_systems

    def get_smart_targets(
        self,
        subject: DiagramBuilder,
        targets: List[DiagramTarget],
    ) -> List[DiagramTarget]:
        """
        Description:
            For each target with None ports, we will try to
            "smartly" create the targets that we want to monitor.
        :param subject:
        :param targets:
        :return:
        """
        # Setup
        smart_targets = []

        for target in targets:
            if target.ports is not None:
                smart_targets.append(target)
                continue

            # If ports is None, then we will try to "smartly" create the ports
            system = subject.GetSubsystemByName(target.name)

            # TODO: Add support for investigating output ports
            num_ports = system.num_output_ports()
            if num_ports == 0:
                loguru.logger.warning(f"System {target.name} has no output ports! Skipping...")
                continue

            output_ports_to_watch = [port_index for port_index in range(num_ports)]

            # Add a target with the connected ports to the smart_targets list
            smart_targets.append(
                DiagramTarget(target.name, output_ports_to_watch),
            )

        # Log the list of inferred targets
        loguru.logger.info(f"Found {len(smart_targets)} inferred targets:")
        for idx, target in enumerate(smart_targets):
            loguru.logger.info(f"{idx}: {target.name} - {target.ports}")

        return smart_targets

    def save_figures(self):
        """
        Description:
            Saves all the figures from the port watchers.
        :return:
        """
        for system_name in self.port_watchers:
            system_ii = self.diagram.GetSubsystemByName(system_name)
            ports_on_ii = self.port_watchers[system_name]
            for port_name in ports_on_ii:
                temp_port_watcher = ports_on_ii[port_name]
                temp_port_watcher.save_figures(self.diagram_context)



from .search import (
    find_all_systems_with_input_port,
    find_all_systems_with_output_port,
)
from .constants import Performer

__all__ = [
    "find_all_systems_with_input_port",
    "find_all_systems_with_output_port",
    "Performer",
]


import os
from pathlib import Path
from typing import Tuple, Callable

import networkx as nx
import numpy as np
from pydrake.common.value import AbstractValue
from pydrake.math import RigidTransform
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Diagram, Context
from pydrake.systems.primitives import ConstantVectorSource, ConstantValueSource

from brom_drake.all import add_watcher_and_build
from brom_drake.motion_planning.systems.prototypical_planner import PrototypicalPlannerSystem
from brom_drake.scenes.types import BaseScene
from brom_drake.scenes.roles import kOfflineMotionPlanner, Role
from brom_drake.file_manipulation.urdf.simple_shape_urdfs.shape_definition import SphereDefinition
from brom_drake.file_manipulation.urdf.simple_shape_urdfs.urdf_definition import SimpleShapeURDFDefinition
from brom_drake.utils import Performer


class OfflineMotionPlanningScene(BaseScene):
    def __init__(
        self,
        time_step: float = 1e-3,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def add_all_secondary_cast_members_to_builder(self):
        pass

    def add_start_and_goal_to_this_plant(self, plant: MultibodyPlant):
        """
        Description
        -----------
        Add the start and goal to this plant.
        Parameters
        ----------
        :param plant: The plant to add the start and goal to.

        Notes
        -----
        Feel free to overwrite this if you have specific sizes of the start and goal in mind,
        or if you want to add more than just spheres, etc.
        :return:

        """

        # Create a sphere for the start
        start_sphere_defn = SimpleShapeURDFDefinition(
            name="start_sphere",
            shape=SphereDefinition(radius=0.1),
            create_collision=False,
            color=np.array([0, 0.3, 0.8, 0.3]),
        )
        start_sphere_urdf_location = Path("brom/models/spheres/start.urdf")
        os.makedirs(start_sphere_urdf_location.parent, exist_ok=True)

        start_sphere_defn.write_to_file(start_sphere_urdf_location)

        # Load the start sphere into the plant and rigidly attach it at the start_pose
        start_sphere_model_idx = Parser(plant).AddModels(str(start_sphere_urdf_location))[0]
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("start_sphere_base_link", start_sphere_model_idx),
            self.start_pose,
        )

        # Create a sphere for the goal
        goal_sphere_defn = SimpleShapeURDFDefinition(
            name="end_sphere",
            shape=SphereDefinition(radius=0.1),
            create_collision=False,
            color=np.array([0, 0.8, 0.3, 0.3]),
        )
        goal_sphere_urdf_location = Path("brom/models/spheres/end.urdf")
        os.makedirs(goal_sphere_urdf_location.parent, exist_ok=True)

        goal_sphere_defn.write_to_file(goal_sphere_urdf_location)

        # Load the start sphere into the plant and rigidly attach it at the start_pose
        goal_sphere_model_idx = Parser(plant).AddModels(str(goal_sphere_urdf_location))[0]
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName(f"{goal_sphere_defn.name}_base_link", goal_sphere_model_idx),
            self.goal_pose,
        )

    def add_start_source_system(self):
        """
        Add a source system for the start pose.
        :return:
        """
        # Setup
        X_start = self.start_pose
        p_start = np.hstack((
            X_start.translation(),
            X_start.rotation().ToQuaternion().wxyz()
        ))

        # Adds a source system for the start pose to the builder
        start_source_system = ConstantVectorSource(p_start)
        start_source_system.set_name("start_pose")

        self.builder.AddSystem(start_source_system)

    def add_goal_source_system(self):
        """
        Add a source system for the goal pose.
        :return:
        """
        # Setup
        X_goal = self.goal_pose
        p_goal = np.hstack((
            X_goal.translation(),
            X_goal.rotation().ToQuaternion().wxyz()
        ))

        # Adds a source system for the goal pose to the builder
        goal_source_system = ConstantVectorSource(p_goal)
        goal_source_system.set_name("goal_pose")

        self.builder.AddSystem(goal_source_system)

    def cast_scene(
        self,
        cast: Tuple[Role, Performer] = [],
    ):
        """
        Description
        -----------
        This small modification to the normal cast_scene()
        function includes a call to create the optional inputs if needed.
        :param cast:
        :return:
        """
        # Setup

        # Default
        super().cast_scene(cast=cast)

        # Create optional outputs, if needed
        for role_ii, performer_ii in cast:
            self.create_optional_outputs_if_necessary(
                role_ii, performer_ii
            )

    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[nx.DiGraph, np.ndarray],
        ],
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        This function is used to easily cast and build the scene.
        :param planning_algorithm: The algorithm that we will use to
        :return:
        """
        # Setup

        # Add all elements to the builder
        self.add_all_secondary_cast_members_to_builder()

        # Create a planner from the algorithm
        prototypical_planner = PrototypicalPlannerSystem(
            self.plant, self.scene_graph,
            planning_algorithm,
        )

        # Cast the scene using the prototypical planner
        planner_role = self.suggested_roles()[0]

        # Fulfill each role-performer pair in the casting_call list
        self.fill_role(planner_role, prototypical_planner)

        self.create_optional_outputs_if_necessary(
            planner_role, prototypical_planner
        )

        # Build
        # self.diagram = builder.Build()
        # self.diagram_context = self.diagram.CreateDefaultContext()

        watcher, self.diagram, self.diagram_context = add_watcher_and_build(self.builder)

        return self.diagram, self.diagram_context

    def create_optional_outputs_if_necessary(
        self,
        role: Role, # This role will always be the Motion Planner, won't it?
        performer: Performer,
    ):
        """
        Description
        -----------
        This method checks to see if all optional outputs were connected
        for the given role. If all of them were, then we do nothing.
        If any of the optional outputs were NOT connected,
        then we add a dummy value to replace it.

        Requirements
        ------------
        This method should only be called after the full scene has
        been built. Otherwise, this may have unexpected behavior.

        :return:
        """
        # Setup

        # Iterate through all OPTIONAL OUTPUTS
        # for assignment_ii in role.port_assignments:
        #     # Should we investigate this port?
        #     investigate_this_assignment = assignment_ii.pairing_type == PairingType.kOutput
        #     investigate_this_assignment = investigate_this_assignment and \
        #                                   (not assignment_ii.is_required)
        #
        #     if not investigate_this_assignment:
        #         continue # Skip this assignment value if it doesn't satisfy this
        #
        #     # Otherwise, check to see if we have the desired output port
        #     if performer.HasOutputPort(assignment_ii.performer_port_name):
        #         continue # If performer has the output port, then we should have already connected it; skip.
        #
        #     # If performer DOES NOT HAVE

        # TODO(kwesi): Maybe move above, general code to somewhere else?
        if role.name != "OfflineMotionPlanner":
            raise ValueError(
                f"Expected role for OfflineMotionPlanning scene to be \"OfflineMotionPlanner\";" +
                f"received {role.name}"
            )

        # Check to see if performer has the LAST assignment in the motion planning role
        last_assignment = role.port_assignments[-1]
        if last_assignment.performer_port_name != "plan_is_ready":
            raise ValueError(
                f"Expected last port assignment to be \"plan_is_ready\" assignment, but" +
                f"received an assignment with performer_port_name \"{last_assignment.performer_port_name}\""
            )

        if performer.HasOutputPort(last_assignment.performer_port_name):
            return # Do nothing; performer should already be connected

        # If the performer does not have plan_is_ready port, then
        # let's create a dummy value and connect it to the right place.
        plan_ready_source = ConstantValueSource(
            AbstractValue.Make(True),
        )
        self.builder.AddSystem(plan_ready_source)

        # Find system we want to connect it to
        systems_list = last_assignment.find_any_matching_input_targets(self.builder)
        assert len(systems_list) == 1, \
            f"Expected 1 system to have port \"{last_assignment.external_target_name}\"," + \
            f" but found {len(systems_list)} systems with that output port."
        # TODO(kwesi): Perhaps move more of these error assertions to a separate file?

        self.builder.Connect(
            plan_ready_source.get_output_port(),
            systems_list[0].GetInputPort(last_assignment.external_target_name)
        )

    @property
    def start_pose(self) -> RigidTransform:
        """
        Get the start pose. This should be defined by the subclass.
        :return:
        """
        return None

    @property
    def goal_pose(self) -> RigidTransform:
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        return None

    def suggested_roles(self):
        return [kOfflineMotionPlanner]


"""
errors.py
Description:

    Contains the custom errors for the DiagramWatcher module.
"""
from typing import List

from .constants import INELIGIBLE_SYSTEM_TYPES
from brom_drake import DiagramTarget


class UnrecognizedTargetError(ValueError):
    """
    UnrecognizedSystemError
    Description:

        Raised when a system is not recognized by the DiagramWatcher.
    """
    def __init__(self, target: DiagramTarget, system_names: List[str] = None):
        self.target = target
        self.message = f"Target with name {target.name} is not recognized by the DiagramWatcher.\n"
        if system_names is not None:
            self.message += f"(Available systems are {system_names})\n"

        self.message += f"If system with name {target.name} does exist, it may be one of the ineligible types:\n "
        self.message += f"\t{INELIGIBLE_SYSTEM_TYPES}.\n"

        super().__init__(self.message)




from setuptools import find_packages, setup
import codecs
import os

if __name__ == "__main__":
    here = os.path.abspath(os.path.dirname(__file__))

    with codecs.open(
        os.path.join(here, "README.md"), encoding="utf-8"
    ) as f:
        long_description = "\n" + f.read()

    setup(
        name="brom_drake",
        version='0.2.7',
        author="Kwesi Rutledge",
        author_email="thesolitaryecrivain@gmail.com",
        url="https://github.com/kwesiRutledge/brom_drake-py",
        description="A set of convenient logging and testing tools for the Drake robotics toolbox.",
        long_description_content_type="text/markdown",
        long_description=long_description,
        # packages=find_packages(where='src/brom_drake'),
        install_requires=[
            'drake', 'meshcat', 'manipulation',
            'loguru', 'matplotlib', 'numpy',
            'pycollada', 'trimesh',
        ],
        keywords=["drake", "robotics", "testing", "logging"],
        classifiers=[
            "Programming Language :: Python :: 3",
        ]
    )


"""
rrt_plan_generator.py
Description:

    This file defines the RRT planner class, which is used to plan motion for a robot using the Rapidly-exploring Random Tree (RRT) algorithm.
"""
import networkx as nx
import numpy as np
from pydrake.common.eigen_geometry import Quaternion
from pydrake.common.value import AbstractValue
from pydrake.geometry import SceneGraph
from pydrake.math import RigidTransform, RotationMatrix
from pydrake.solvers import OsqpSolver, MathematicalProgram, Solve, SolutionResult
from pydrake.multibody.inverse_kinematics import DoDifferentialInverseKinematics, \
    DifferentialInverseKinematicsParameters, InverseKinematics
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.systems.framework import LeafSystem, BasicVector, Context

from brom_drake.motion_planning.algorithms.rrt.base import BaseRRTPlanner, BaseRRTPlannerConfig


class RRTPlanGenerator(LeafSystem):
    """
    RRTPlanGenerator class
    Description:
        This class will call the RRT planning algorithm for
        and then share it in an output port.
    """
    def __init__(
        self,
        plant: MultibodyPlant,
        scene_graph: SceneGraph,
        robot_model_idx: ModelInstanceIndex = None,
        rrt_config: BaseRRTPlannerConfig = None
    ):
        """
        Description:
            This function initializes the RRT planner.
        """
        super().__init__()
        # Setup
        self.robot_model_idx = robot_model_idx
        self.plant, self.scene_graph = plant, scene_graph
        self.plan = None
        self.rrt_config = rrt_config

        # Compute dof using robot_model_idx
        self.dim_q = None

        # Set Up Input and Output Ports
        self.root_context = None # NOTE: This should be assigned by the user before calling!
        self.create_input_ports()
        self.create_output_ports()

    def compute_plan_if_not_available(self, context, output: AbstractValue):
        """
        Description:
            This function computes the plan if it is not available.
        """
        # Print
        if self.plan is None:
            # Compute plan
            p_WStart_vec = self.GetInputPort("start_pose").Eval(context)
            p_WGoal_vec = self.GetInputPort("goal_pose").Eval(context)
            self.robot_model_idx = self.GetInputPort("robot_model_index").EvalAbstract(context).get_value()

            print(self.robot_model_idx)

            q_start = self.solve_pose_ik_problem(
                p_WStart_vec,
            )

            q_goal = self.solve_pose_ik_problem(
                p_WGoal_vec,
            )

            base_rrt = BaseRRTPlanner(
                self.robot_model_idx,
                self.plant, self.scene_graph,
                config=self.rrt_config,
            )

            if self.root_context is None:
                raise ValueError("Plant context is not initialized yet!")

            base_rrt.root_context = self.root_context

            # Plan and extract path
            rrt, found_path = base_rrt.plan(
                q_start, q_goal,
            )

            if not found_path:
                raise RuntimeError("No path found! Try increasing the number of iterations or checking your problem!")

            path = nx.shortest_path(rrt, source=0, target=rrt.number_of_nodes()-1)
            self.plan = np.array([rrt.nodes[node]['q'] for node in path])

        output.SetFrom(
            AbstractValue.Make(self.plan)
        )

    def create_input_ports(self):
        """
        Description:
            This function creates the input ports for the RRT planner.
        """
        self.DeclareVectorInputPort(
            "start_pose",
            BasicVector(np.zeros((7,))),
        )
        self.DeclareVectorInputPort(
            "goal_pose",
            BasicVector(np.zeros((7,))),
        )
        self.DeclareAbstractInputPort(
            "robot_model_index",
            AbstractValue.Make(ModelInstanceIndex(-1))
        )

    def create_output_ports(self):
        """
        Description:
            This function creates the output ports for the RRT planner.
        """
        # Setup

        # Create output for plan
        sample_plan = np.zeros((0, self.n_actuated_dof))
        self.DeclareAbstractOutputPort(
            "motion_plan",
            lambda: AbstractValue.Make(sample_plan),
            self.compute_plan_if_not_available
        )

        # Create output for plan readiness
        self.DeclareAbstractOutputPort(
            "plan_is_ready",
            lambda: AbstractValue.Make(False),
            self.GetPlanIsReady,
        )

    def define_pose_ik_problem(
        self,
        input_pose_vec: np.ndarray,
        eps0: float = 2.5e-2,
    ) -> InverseKinematics:
        """
        Description
        -----------
        Sets up the inverse kinematics problem for the start pose
        input to theis function.
        :return:
        """
        # Setup
        # p_WStart = RigidTransform(
        #     Quaternion(p_WStart_vec[3:]),
        #     p_WStart_vec[:3]
        # )

        # Create IK Problem
        ik_problem = InverseKinematics(self.plant)

        # Add Pose Target
        ik_problem.AddPositionConstraint(
            self.plant.world_frame(),
            input_pose_vec[:3],
            self.plant.GetFrameByName("ft_frame"),
            (- np.ones((3,)) * eps0).reshape((-1, 1)),
            (+ np.ones((3,)) * eps0).reshape((-1, 1)),
        )

        # TODO(kwesi): Add OrientationCosntraint
        # ik_problem.AddOrientationConstraint(
        #     self.plant.world_frame(),
        #     RotationMatrix(Quaternion(input_pose_vec[3:]).rotation()),
        #     self.plant.GetFrameByName("ft_frame"),
        #     RotationMatrix.Identity(),
        #     0.25,
        # )

        ik_problem.AddOrientationCost(
            self.plant.world_frame(),
            RotationMatrix(Quaternion(input_pose_vec[3:]).rotation()),
            self.plant.GetFrameByName("ft_frame"),
            RotationMatrix.Identity(),
            0.25,
        )

        return ik_problem

    def solve_pose_ik_problem(
        self,
        input_pose_vec: np.ndarray
    ) -> np.ndarray:
        """
        Description:
        :param input_pose_vec:
        :return:
        """
        # Setup

        # Define Problem
        ik_problem = self.define_pose_ik_problem(input_pose_vec)

        # Solve problem
        ik_program = ik_problem.prog()
        ik_result = Solve(ik_program)

        assert ik_result.get_solution_result() == SolutionResult.kSolutionFound, \
            f"Solution result was {ik_result.get_solution_result()}; need SolutionResult.kSolutionFound to make RRT Plan!"

        q_solution = ik_result.get_x_val()

        # Extract only the positions that correspond to our robot's joints
        robot_joint_names = self.plant.GetPositionNames(self.robot_model_idx, add_model_instance_prefix=True)
        all_joint_names = self.plant.GetPositionNames()
        q_out_list = []
        for ii, joint_name in enumerate(all_joint_names):
            if joint_name in robot_joint_names:
                q_out_list.append(q_solution[ii])

        return np.array(q_out_list)


    def set_dimension(self, dim_q: int):
        """
        Description:
            This function sets the dimension of the configuration space.
        """
        self.dim_q = dim_q

    @property
    def n_actuated_dof(self) -> int:
        """
        Description
        -----------
        This method uses the plant and the robot model idx
        to identify the numer of degrees of freedom we have for control.
        :return:
        """
        return self.plant.num_actuated_dofs()

    def GetPlanIsReady(self, context, output: AbstractValue):
        """
        Description:
            This function checks if the plan is ready.
        """
        output.SetFrom(
            AbstractValue.Make(self.plan is not None)
        )

    def set_internal_root_context(self, root_context_in: Context):
        self.root_context = root_context_in


from typing import List, Tuple

import numpy as np
from manipulation.scenarios import AddMultibodyTriad
from pydrake.geometry import Meshcat, MeshcatVisualizer
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph
from pydrake.systems.framework import DiagramBuilder, Diagram, Context

from pydrake.systems.primitives import ConstantVectorSource, VectorLogSink

# Internal Imports
from brom_drake.robots import find_base_link_name_in
from brom_drake.scenes.types import BaseScene
from brom_drake.scenes import SceneID
from brom_drake.scenes.roles.role import Role
from brom_drake.utils import Performer
from .show_me_system import ShowMeSystem

class ShowMeThisModel(BaseScene):
    def __init__(
        self,
        path_to_model: str,
        with_these_joint_positions: List[float] = None,
        base_link_name: str = None,
        time_step: float = 1e3,
        meshcat_port_number: int = 7001, # Usually turn off for CI (i.e., make it None)
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Add the model to the scene
        self.path_to_model = path_to_model
        self.q_des = with_these_joint_positions
        self.base_link_name = base_link_name
        self.time_step = time_step
        self.meshcat_port_number = meshcat_port_number

        # If the base link name is not provided,
        # then we will try to do a smart search for it later.
        self.plant, self.scene_graph = AddMultibodyPlantSceneGraph(
            self.builder,
            time_step=self.time_step,
        )
        self.plant.set_name("ShowMeThisModel_plant")

        self.meshcat = None
        self.model_index, self.model_name = None, None
        self.show_me_system = None

    def add_all_secondary_cast_members_to_builder(self):
        """
        Description
        -----------
        This method will add just the user's model to the builder.

        :return:
        """
        # Setup

        # Add Model
        model_idcs = Parser(plant=self.plant).AddModels(self.path_to_model)
        assert len(model_idcs) == 1, \
            f"Only one model should be added in this scene;" + \
            f" received {len(model_idcs)} in the file {self.path_to_model}."
        self.model_index = model_idcs[0]
        self.model_name = self.plant.GetModelInstanceName(model_idcs[0])

        # Collect the expected number of actuated joints
        n_dofs = self.plant.num_actuated_dofs()
        if self.q_des is None:
            self.q_des = np.zeros(n_dofs)
        else:
            assert len(self.q_des) == n_dofs, \
                f"Expected {n_dofs} joint positions; received {len(self.q_des)}."

        # Add ShowMeSystem
        self.show_me_system = self.builder.AddSystem(
            ShowMeSystem(
                plant=self.plant,
                model_index=self.model_index,
                desired_joint_positions=np.array(self.q_des),
            ),
        )

        # Add Source
        desired_joint_positions_source = self.builder.AddSystem(
            ConstantVectorSource(np.array(self.q_des)),
        )

        # Connect the source to the system
        self.builder.Connect(
            desired_joint_positions_source.get_output_port(0),
            self.show_me_system.get_input_port(0),
        )

        # Add Sink
        output_joints_sink = self.builder.AddSystem(
            VectorLogSink(len(self.q_des)),
        )

        # Connect the system to the sink
        self.builder.Connect(
            self.show_me_system.get_output_port(0),
            output_joints_sink.get_input_port(0),
        )

        # Add A Triad to base?
        AddMultibodyTriad(self.plant.world_frame(), self.scene_graph)

        # Try to collect the base link name if it is not provided
        if self.base_link_name is None:
            self.base_link_name = find_base_link_name_in(self.path_to_model)

        # Weld the base link to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName(self.base_link_name, self.model_index),
        )

        # Connect to Meshcat, if requested
        if self.meshcat_port_number is not None:
            self.meshcat = Meshcat(port=self.meshcat_port_number)  # Object provides an interface to Meshcat
            m_visualizer = MeshcatVisualizer(self.meshcat)
            m_visualizer.AddToBuilder(self.builder, self.scene_graph, self.meshcat)

        # Finalize plant and connect it to system
        self.plant.Finalize()

    def cast_scene_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
    ) -> Tuple[Diagram, Context]:
        super().cast_scene_and_build(cast)

        # Assign the diagram context to the internal show_me_system
        self.show_me_system.mutable_plant_context = self.plant.GetMyMutableContextFromRoot(
            self.diagram_context,
        )

        return self.diagram, self.diagram_context

    @property
    def id(self):
        return SceneID.kShowMeThisModel

    @property
    def suggested_roles(self) -> List[Role]:
        return []



from importlib import resources as impresources
from pathlib import Path

import numpy as np
from pydrake.geometry import SceneGraph, Meshcat, MeshcatVisualizer
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import FixedOffsetFrame
from pydrake.systems.framework import Diagram, DiagramBuilder
from pydrake.systems.primitives import Demultiplexer

from brom_drake.control import IdealJointPositionController
# Local imports
from brom_drake.robots.gripper_type import GripperType
from brom_drake.control.grippers.gripper_controller import GripperController
from brom_drake.file_manipulation.urdf import DrakeReadyURDFConverter

from brom_drake import robots

class UR10eStation(Diagram):
    """
    A template system diagram for controlling a UR10e robot in a simulated environment.
    """

    def __init__(
        self,
        time_step: float = 0.002,
        gripper_type: GripperType = GripperType.NoGripper,
        force_conversion_of_original_urdf: bool = False,
        meshcat_port_number: int = None,
    ):
        """
        Description:

            This class defines the KINEMATIC UR10e Manipulation Station.
            This station is based off of the Drake Manipulation Station showcased
            in Russ Tedrake's manipulation course.
        """
        # Input Processing
        self.force_conversion_of_original_urdf = force_conversion_of_original_urdf
        self.meshcat_port_number = meshcat_port_number

        # Initialize the Diagram
        Diagram.__init__(self)
        self.set_name("UR10e_Station")

        # Create a diagram builder
        self.builder = DiagramBuilder()

        # Create scene_graph and plant
        self.plant, self.scene_graph = None, None
        self.time_step = time_step
        self.create_plant_and_scene_graph()

        # Body ID's and Poses for Anything else in the scene
        self.object_ids = []
        self.object_poses = []

        # Whether we have a camera in the simulation
        # self.has_camera = False
        self.arm = None # Assign the Arm's Model Index later
        self.AddArm()

        # Which sort of gripper we're using (if any)
        self.gripper_type = gripper_type
        if gripper_type == GripperType.Robotiq_2f_85:
            self.Add2f85Gripper()

        # Visualization
        self.meshcat = None


    def AddArm(self):
        """
        Add the UR10e arm to the system.
        """
        # Setup
        original_arm_urdf_path = str(
            impresources.files(robots) / "models/ur/ur10e.urdf",
        )
        expected_arm_urdf_path = Path("./brom/models/ur10e/ur10e.drake.urdf")

        # Convert the arm urdf if necessary
        arm_urdf = None
        if (not expected_arm_urdf_path.exists()) or self.force_conversion_of_original_urdf:
            # If drake-compatible URDF does not exist, then we need to convert
            # the original URDF to a drake-compatible URDF.
            arm_urdf = DrakeReadyURDFConverter(original_arm_urdf_path).convert_urdf()
            arm_urdf = str(arm_urdf)
        else:
            # Otherwise, let's just read the drake-compatible URDF
            arm_urdf = str(expected_arm_urdf_path)

        self.end_effector_frame_name = "tool0"

        # The hardware system has lots of damping so this is more realistic,
        # but requires a simulation with small timesteps.

        self.arm = Parser(plant=self.plant).AddModels(arm_urdf)[0]

        # Fix the base of the arm to the world
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("base_link", self.arm),
        )

        # Create a new frame with the actual end-effector position.
        self.X_ee = RigidTransform()
        self.X_ee.set_translation([0, 0, 0.13])
        self.plant.AddFrame(FixedOffsetFrame(
            "end_effector",
            self.plant.GetFrameByName("tool0"),
            self.X_ee, self.arm))

    def Add2f85Gripper(self):
        """
        Add the Robotiq 2F-85 gripper to the system. The arm must be added first.
        """
        # Setup

        # Add a gripper with actuation to the full simulated plant
        gripper_urdf_path = str(
            impresources.files(robots) / "models/2f_85_gripper/urdf/robotiq_2f_85.urdf"
        )
        self.gripper = Parser(plant=self.plant).AddModels(gripper_urdf_path)[0]

        X_grip = RigidTransform()
        X_grip.set_rotation(
            RotationMatrix(RollPitchYaw([0.0, 0.0, np.pi / 2]))
        )
        self.plant.WeldFrames(
            self.plant.GetFrameByName("tool0", self.arm),
            self.plant.GetFrameByName("robotiq_arg2f_base_link", self.gripper),
            X_grip,
        )

        # Add a gripper without actuation to the controller plant
        gripper_static_urdf = str(
            impresources.files(robots) / "models/2f_85_gripper/urdf/robotiq_2f_85_static.urdf"
        )
        static_gripper = Parser(plant=self.controller_plant).AddModels(
            gripper_static_urdf
        )[0]

        self.controller_plant.WeldFrames(
            self.controller_plant.GetFrameByName("tool0", self.controller_arm),
            self.controller_plant.GetFrameByName("robotiq_arg2f_base_link", static_gripper),
            X_grip,
        )

    def ConnectToMeshcatVisualizer(self, port=None):
        self.meshcat = Meshcat(port)
        m = MeshcatVisualizer(self.meshcat)
        m.AddToBuilder(self.builder, self.scene_graph, self.meshcat)

        print("Open %s in a browser to view the meshcat visualizer." % self.meshcat.web_url())

    def create_plant_and_scene_graph(self):
        """
        Description
        -----------
        This function creates the plant for the UR10e station, if needed.
        :return:
        """
        # Setup

        # Create SceneGraph
        self.scene_graph = self.builder.AddSystem(
            SceneGraph()
        )
        self.scene_graph.set_name(f"{self.get_name()}_SceneGraph")

        # Create plant
        self.plant = self.builder.AddSystem(
            MultibodyPlant(time_step=self.time_step)
        )
        self.plant.RegisterAsSourceForSceneGraph(self.scene_graph)
        self.plant.set_name(f"{self.get_name()}_Plant")

    def Finalize(self):
        """
        Description
        ----------
        This finalizes the diagram by:
        - Finalizing all plants in the diagram
        :return: Nothing
        """
        # Setup

        # Finalize all plants
        self.plant.Finalize()

        # Set up the scene graph
        self.builder.Connect(
            self.scene_graph.get_query_output_port(),
            self.plant.get_geometry_query_input_port()
        )

        self.builder.Connect(
            self.plant.get_geometry_pose_output_port(),
            self.scene_graph.get_source_pose_port(self.plant.get_source_id())
        )

        self.builder.ExportOutput(
            self.scene_graph.get_query_output_port(),
            "query_object",
        )

        # Connect Meshcat, if desired
        if self.use_meshcat():
            self.ConnectToMeshcatVisualizer(port=self.meshcat_port_number)

        # Create Arm and Gripper Controllers
        self.CreateArmPorts()

        if self.gripper_type != GripperType.NoGripper:
            self.CreateGripperControllerAndConnect()

        # Build the diagram
        self.builder.BuildInto(self)

    def CreateArmPorts(
        self,
    ):
        """
        Description
        -----------
        This function creates a Cartesian arm controller and connects it to the rest of the system.
        :return:
        """
        # Setup

        # Create the input ports for the joint positions
        self.arm_controller = self.CreateJointArmControllerAndExportIO()

        # Output measured arm position and velocity
        demux = self.builder.AddSystem(
            Demultiplexer(self.plant.num_multibody_states(self.arm), self.plant.num_positions(self.arm)),
        )
        demux.set_name(f"{self.get_name()}_Demux")

        self.builder.Connect(
            self.plant.get_state_output_port(self.arm),
            demux.get_input_port(0),
        )
        self.builder.ExportOutput(
            demux.get_output_port(0),
            "measured_arm_position",
        )
        self.builder.ExportOutput(
            demux.get_output_port(1),
            "measured_arm_velocity",
        )

    def CreateJointArmControllerAndExportIO(
        self,
    )-> IdealJointPositionController:
        """
        Description
        -----------
        Creates an ideal joint position controller and exports the necessary inputs + outputs.
        :return:
        """
        # Setup
        arm = self.arm
        plant = self.plant

        # Create the joint controller
        joint_controller = self.builder.AddSystem(
            IdealJointPositionController(arm, plant),
        )

        # Export the input and output ports
        self.builder.ExportInput(
            joint_controller.desired_joint_positions_port,
            "desired_joint_positions",
        )

        self.builder.ExportOutput(
            joint_controller.GetOutputPort("measured_joint_positions"),
            "measured_joint_positions",
        )

        return joint_controller



    def CreateGripperControllerAndConnect(self):
        """
        Description
        -----------
        This function creates a gripper controller and connects it to the rest of the system.
        Requires that the builder has been created and IS NOT FINALIZED.
        :return: (Nothing)
        """
        # Setup

        # Create gripper controller
        gripper_controller = self.builder.AddSystem(
            GripperController(self.gripper_type)
        )
        gripper_controller.set_name("gripper_controller")

        # Connect gripper controller to the diagram
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target"),
            "gripper_target")
        self.builder.ExportInput(
            gripper_controller.GetInputPort("gripper_target_type"),
            "gripper_target_type")

        self.builder.Connect(
            self.plant.get_state_output_port(self.gripper),
            gripper_controller.GetInputPort("gripper_state"))
        self.builder.Connect(
            gripper_controller.GetOutputPort("applied_gripper_torque"),
            self.plant.get_actuation_input_port(self.gripper))

        # Send gripper position and velocity as an output
        self.builder.ExportOutput(
            gripper_controller.GetOutputPort("measured_gripper_position"),
            "measured_gripper_position")
        self.builder.ExportOutput(
            gripper_controller.GetOutputPort("measured_gripper_velocity"),
            "measured_gripper_velocity")

    def use_meshcat(self):
        return self.meshcat_port_number is not None


from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.math import SpatialVelocity
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant, CoulombFriction
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.systems.framework import LeafSystem, PortDataType, BasicVector, Context

import numpy as np

# Internal Imports
from brom_drake.robots.utils import find_base_link_name_in


class IdealJointPositionController(LeafSystem):
    def __init__(
        self,
        robot_model: ModelInstanceIndex,
        plant: MultibodyPlant,
        q0: np.ndarray = None,
        plant_context: Context = None,
    ):
        LeafSystem.__init__(self)
        # Input Processing

        # Set Up
        self.plant, self.robot_model = plant, robot_model
        self.plant_context = plant_context

        self.q0 = q0

        # Create Input Port for the Body's Joint Positions
        self.desired_joint_positions_port = self.DeclareVectorInputPort(
            "desired_joint_positions",
            BasicVector(self.plant.num_actuated_dofs()),
        )

        # Create Output Port which should share the pose of the block
        self.DeclareVectorOutputPort(
            "measured_joint_positions",
            BasicVector(self.plant.num_actuated_dofs()),
            self.SetJointPositions,
            {self.time_ticket()}    # indicate that this doesn't depend on any inputs,
        )                           # but should still be updated each timestep


    def SetJointPositions(self, context, output):
        """
        Description:
            This function sets the desired pose of the block.
        """
        # Setup
        plant_context = self.plant_context
        robot_model = self.robot_model

        # Get Desired Pose from Port
        joint_positions = self.desired_joint_positions_port.Eval(context)

        # Get plant context from station context
        # station_context = self.station_context
        # if station_context is None:
        #     raise RuntimeError("station_context is None")

        self.plant.SetPositions(plant_context, robot_model, joint_positions)
        self.plant.SetVelocities(
            plant_context,
            robot_model,
            np.zeros(self.plant.num_velocities(robot_model))
        )

        output.SetFromVector(joint_positions)

    def SetInitialJointPositions(self, diagram_context):
        """
        Description:
            Sets the initial position .
        """
        # Setup
        q0 = self.q0
        if q0 is None:
            q0 = np.zeros(self.plant.num_positions())

        station_context = self.station_context
        if station_context is None:
            station_context = diagram_context
        plant_context = self.GetSubsystemContext(self.plant, station_context)

        # Set Joint Positions and velocities
        self.plant.SetPositions(plant_context, self.robot_model, q0)

        self.plant.SetVelocities(
            plant_context,
            self.robot_model,
            np.zeros(self.plant.num_velocities(self.robot_model))
        )